#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Related scripts for the HLI-STR (TREDPARSE) paper.
"""
import os.path as op
import os
import csv
import sys
import vcf
import logging
import shutil
import json
import numpy as np
import pandas as pd

from collections import defaultdict
from random import sample
from pyfaidx import Fasta
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from itertools import product
from natsort import natsorted

from jcvi.graphics.base import (
    FancyArrow,
    normalize_axes,
    panel_labels,
    plt,
    savefig,
    set_helvetica_axis,
)
from jcvi.formats.base import is_number, must_open
from jcvi.formats.sam import get_minibam_bed, index
from jcvi.variation.str import TREDsRepo, af_to_counts, read_treds
from jcvi.utils.cbook import percentage
from jcvi.utils.table import tabulate
from jcvi.apps.grid import Parallel
from jcvi.apps.bwa import align
from jcvi.apps.base import datafile, sh
from jcvi.assembly.sim import eagle, wgsim
from jcvi.apps.base import OptionParser, ActionDispatcher, mkdir, iglob


# Huntington risk allele
infected_thr = 40
ref_thr = 19
SIMULATED_HAPLOID = r"Simulated haploid $\mathit{h}$"
SIMULATED_DIPLOID = r"Simulated diploid $\mathit{20/h}$"
lsg = "lightslategray"

# List of TRED loci excluded from plots
ignore = ("AR",)


class TREDPARSEvcf(object):
    def __init__(self, vcffile):
        samplekey = op.basename(vcffile).split(".")[0]
        reader = vcf.Reader(open(vcffile, "rb"))
        res = "-1/-1"
        for rec in reader:
            sample = rec.samples[0]
            res = sample["GB"]
            ci = sample["CI"]
            break
        print(samplekey, res, ci)


class TrioOrDuo:
    def __init__(self, parents, child, family):
        self.parents = dict((x, family[x]) for x in parents)
        self.child = dict((x, family[x]) for x in child)
        self.is_trio = len(self.parents) == 2

    def __len__(self):
        return len(self.parents) + len(self.child)

    def __key(self):
        return tuple(sorted(self.parents.values()) + self.child.values())

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.__key() == other.__key()

    def __str__(self):
        return str(self.parents) + "=>" + str(self.child)

    __repr__ = __str__

    def check_mendelian(self, df, tred, tolerance=0, x_linked=False, verbose=False):
        child_key = self.child.values()[0]
        c = get_alleles(df, child_key, tred)
        if c is None:
            return 0
        if self.is_trio:
            parent_keys = self.parents.values()
            p1 = get_alleles(df, parent_keys[0], tred)
            p2 = get_alleles(df, parent_keys[1], tred)
            if (p1 is None) or (p2 is None):
                return 0
            possible_progenies = get_progenies(
                p1, p2, x_linked=x_linked, tolerance=tolerance
            )
            mendelian_error = not (c in possible_progenies)
            if verbose:
                print(
                    parent_keys[0],
                    parent_keys[1],
                    child_key,
                    p1,
                    p2,
                    c,
                    not mendelian_error,
                )
        else:
            parent_key = self.parents.values()[0]
            p1 = get_alleles(df, parent_key, tred)
            if p1 is None:
                return 0
            _p1 = expand_alleles(p1, tolerance=tolerance)
            mendelian_error = len(set(_p1) & set(c)) == 0
            if mendelian_error and x_linked:
                # Do not count case where - progeny is male, parent is male
                if (c[0] == c[1]) and (p1[0] == p1[1]):
                    mendelian_error = 0
            if verbose:
                print(parent_key, child_key, p1, c, not mendelian_error)
        return mendelian_error


def expand_alleles(p, tolerance=0):
    """
    Returns expanded allele set given the tolerance.
    """
    _p = set()
    for x in p:
        _p |= set(range(x - tolerance, x + tolerance + 1))
    return _p


def get_progenies(p1, p2, x_linked=False, tolerance=0):
    """
    Returns possible progenies in a trio.
    """
    _p1 = expand_alleles(p1, tolerance=tolerance)
    _p2 = expand_alleles(p2, tolerance=tolerance)
    possible_progenies = set(tuple(sorted(x)) for x in product(_p1, _p2))
    if x_linked:  # Add all hemizygotes
        possible_progenies |= set((x, x) for x in (set(_p1) | set(_p2)))
    return possible_progenies


def get_alleles(df, sample, tred):
    try:
        s = df.ix[sample]
        a = int(s[tred + ".1"])
        b = int(s[tred + ".2"])
    except:
        return None
    if a == -1 or b == -1:
        return None
    return a, b


def main():

    actions = (
        # Prepare data
        ("simulate", "simulate bams with varying inserts with dwgsim"),
        ("mergebam", "merge sets of BAMs to make diploid"),
        ("mini", "prepare mini-BAMs that contain only the STR loci"),
        ("alts", "build alternative loci based on simulation data"),
        # Compile results
        ("batchlobstr", "run lobSTR on a list of BAMs"),
        ("compilevcf", "compile vcf outputs into lists"),
        # Plotting
        ("evidences", "plot distribution of evidences"),
        ("likelihood", "plot likelihood surface"),
        ("likelihood2", "plot likelihood surface and marginals"),
        ("likelihood3", "plot likelihood surface and marginals for two settings"),
        ("compare", "compare callers on fake HD patients"),
        ("compare2", "compare TREDPARSE and lobSTR on fake HD patients"),
        ("power", "compare TREDPARSE on fake HD patients adding evidence"),
        ("tredparse", "compare TREDPARSE on fake HD patients adding coverage"),
        ("allelefreq", "plot the allele frequencies of some STRs"),
        ("allelefreqall", "plot all 30 STR allele frequencies"),
        ("depth", "plot read depths across all TREDs"),
        # Diagram
        ("diagram", "plot the predictive power of various evidences"),
        # Extra analysis for reviews
        ("mendelian", "calculate Mendelian errors based on trios and duos"),
        ("mendelian2", "second iteration of Mendelian error calculation"),
        ("mendelian_errors", "plot Mendelian errors calculated by mendelian"),
        ("mendelian_errors2", "plot Mendelian errors calculated by mendelian2"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def mendelian_errors2(args):
    """
    %prog mendelian_errors2 Trios.summary.csv

    Plot Mendelian errors as calculated by mendelian(). File
    `Trios.summary.csv` looks like:

    Name,Motif,Inheritance,N_Correct,N_Error,N_missing,ErrorRate [N_Error / (N_Correct + N_Error))]
    DM1,CTG,AD,790,12,0,1.5%
    DM2,CCTG,AD,757,45,0,5.6%
    DRPLA,CAG,AD,791,11,0,1.4%
    """
    p = OptionParser(mendelian_errors2.__doc__)
    opts, args, iopts = p.set_image_options(args, figsize="7x7", format="png")

    if len(args) != 1:
        sys.exit(not p.print_help())

    (csvfile,) = args
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])

    ymin = -0.2
    df = pd.read_csv(csvfile)
    data = []
    for i, d in df.iterrows():
        tred = d["Name"]
        motif = d["Motif"]
        if tred in ignore:
            logging.debug("Ignore {}".format(d["TRED"]))
            continue

        if len(motif) > 6:
            if "/" in motif:  # CTG/CAG
                motif = motif.split("/")[0]
            else:
                motif = motif[:6] + ".."
        xtred = "{} {}".format(tred, motif)
        accuracy = d[-1]
        data.append((xtred, accuracy))

    key = lambda x: float(x.rstrip("%"))
    data.sort(key=lambda x: key(x[-1]))
    print(data)
    treds, accuracies = zip(*data)
    ntreds = len(treds)
    ticks = range(ntreds)
    accuracies = [key(x) for x in accuracies]

    for tick, accuracy in zip(ticks, accuracies):
        ax.plot([tick, tick], [ymin, accuracy], "-", lw=2, color="lightslategray")

    (trios,) = ax.plot(accuracies, "o", mfc="w", mec="b")
    ax.set_title("Mendelian errors based on STR calls in trios in HLI samples")
    ntrios = "Mendelian errors in 802 trios"
    ax.legend([trios], [ntrios], loc="best")

    ax.set_xticks(ticks)
    ax.set_xticklabels(treds, rotation=45, ha="right", size=8)
    ax.set_yticklabels([int(x) for x in ax.get_yticks()], family="Helvetica")
    ax.set_ylabel(r"Mendelian errors (\%)")
    ax.set_ylim(ymin, 100)

    normalize_axes(root)

    image_name = "mendelian_errors2." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def mendelian2(args):
    """
    %prog mendelian2
        XC_kinship_TRIO_annotationed_age_sex_PaternalMaternalAgeWhenChildWasBorn.txt
        hli.20170805.tsv

    Second iteration of Mendelian error calculation. This includes all the read
    counts and gender information to correct error estimate of X-linked loci.
    """
    p = OptionParser(mendelian2.__doc__)
    p.add_option(
        "--treds", default=None, help="Extract specific treds, use comma to separate"
    )
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    triofile, hlitsv = args
    repo = TREDsRepo()
    treds = opts.treds.split(",") if opts.treds else repo.names
    triodata = pd.read_csv(triofile, sep="\t")
    samplekey = lambda x: x.split("_")[1]
    trios = []
    for i, row in triodata.iterrows():
        proband = row["proband"]
        parents = row["parents"]
        proband_sex = row["proband_sex"]
        parents_sex = row["parent1_sex,parent2_sex"]
        proband = samplekey(proband)
        p1, p2 = parents.split(",")
        p1, p2 = samplekey(p1), samplekey(p2)
        p1_sex, p2_sex = parents_sex.split(",")
        if p1_sex == "Male":
            p1, p2 = p2, p1
            p1_sex, p2_sex = p2_sex, p1_sex
        trios.append((proband, proband_sex, p1, p1_sex, p2, p2_sex))

    header = "{0}_ID {0}_Sex {0}_Calls"
    header += " {0}_Full {0}_Partial {0}_Repeat {0}_Paired"
    tredsdata = pd.read_csv(hlitsv, sep="\t", low_memory=False)
    tsvfiles = []
    summary = open("Trios.summary.csv", "w")
    summary_header = (
        "Name,Motif,Inheritance,N_Correct,N_Error,N_missing,"
        "ErrorRate [N_Error / (N_Correct + N_Error))]"
    )
    print(summary_header, file=summary)
    print(summary_header)
    for tred in treds:
        if tred in ("FXS", "AR"):
            continue
        tr = repo[tred]
        tsvfile = "{}.details.tsv".format(tred)
        fw = open(tsvfile, "w")
        td = {}
        for _, row in tredsdata.iterrows():
            s = str(row["SampleKey"])
            inferredGender = row["inferredGender"]
            try:
                calls = row[tred + ".calls"]
                fdp = int(row[tred + ".FDP"])
                pdp = int(row[tred + ".PDP"])
                rdp = int(row[tred + ".RDP"])
                pedp = int(row[tred + ".PEDP"])
                td[s] = [str(x) for x in (inferredGender, calls, fdp, pdp, rdp, pedp)]
            except ValueError:
                logging.error("Invalid row: {}".format(row))
                continue

        h = " ".join((header.format("P1"), header.format("P2"), header.format("Kid")))
        print("\t".join(["MendelianError"] + h.split()), file=fw)
        tredcall = lambda x: td.get(x, ["", "-1|-1", "", "", "", ""])[:]
        counts = defaultdict(int)
        is_xlinked = repo[tred].is_xlinked
        shorten = lambda x: str(int(x[-4:]))  # Simplify SampleKey
        for proband, proband_sex, p1, p1_sex, p2, p2_sex in trios:
            tp1 = tredcall(p1)
            tp2 = tredcall(p2)
            tpp = tredcall(proband)
            m = mendelian_check(tp1, tp2, tpp, is_xlinked=is_xlinked)
            counts[m] += 1
            if is_xlinked:
                for (p, p_sex) in ((tp1, p1_sex), (tp2, p2_sex), (tpp, proband_sex)):
                    if p[1].startswith("-"):
                        p[1] = "n.a."
            cells = [shorten(p1), p1_sex] + tp1[1:]
            cells += [shorten(p2), p2_sex] + tp2[1:]
            cells += [shorten(proband), proband_sex] + tpp[1:]
            print("\t".join([m] + cells), file=fw)
        fw.close()
        tsvfiles.append(tsvfile)

        error_rate = counts["Error"] * 100.0 / (counts["Correct"] + counts["Error"])
        line = ",".join(
            str(x)
            for x in (
                tred,
                tr.motif,
                tr.inheritance,
                counts["Correct"],
                counts["Error"],
                counts["Missing"],
                "{:.1f}%".format(error_rate),
            )
        )
        print(line, file=summary)
        print(line)

    # Combine into a master spreadsheet
    import xlwt

    wb = xlwt.Workbook()
    converter = lambda x: int(x) if is_number(x, cast=int) else x
    header = xlwt.easyxf("font: bold on, name Helvetica; align: horiz center")
    hc = "font: name Helvetica; align: horiz center;"
    horiz_center = xlwt.Style.easyxf(hc)
    correct = xlwt.Style.easyxf(hc + "pattern: pattern solid, fore_colour light_green;")
    error = xlwt.Style.easyxf(hc + "pattern: pattern solid, fore_colour rose;")
    missing = xlwt.Style.easyxf(
        hc + "pattern: pattern solid, fore_colour light_yellow;"
    )
    for tsvfile in tsvfiles:
        sheet = op.basename(tsvfile).split(".", 1)[0]
        ws = wb.add_sheet(sheet)
        fp = open(tsvfile, "rb")
        reader = csv.reader(fp, delimiter="\t")
        for r, row in enumerate(reader):
            style = header if r == 0 else horiz_center
            for c, col in enumerate(row):
                if c == 0 and r:
                    style = {"Correct": correct, "Error": error, "Missing": missing}[
                        col
                    ]
            ws.write(r, c, converter(col), style)
            ws.set_panes_frozen(True)
            ws.set_horz_split_pos(1)

    wb.save("Trios.xls")
    summary.close()


def mendelian_check(tp1, tp2, tpp, is_xlinked=False):
    """
    Compare TRED calls for Parent1, Parent2 and Proband.
    """
    call_to_ints = lambda x: tuple(int(_) for _ in x.split("|") if _ != ".")
    tp1_sex, tp1_call = tp1[:2]
    tp2_sex, tp2_call = tp2[:2]
    tpp_sex, tpp_call = tpp[:2]
    # tp1_evidence = sum(int(x) for x in tp1[2:])
    # tp2_evidence = sum(int(x) for x in tp2[2:])
    # tpp_evidence = sum(int(x) for x in tpp[2:])
    tp1_call = call_to_ints(tp1_call)
    tp2_call = call_to_ints(tp2_call)
    tpp_call = call_to_ints(tpp_call)
    possible_progenies = set(tuple(sorted(x)) for x in product(tp1_call, tp2_call))
    if is_xlinked and tpp_sex == "Male":
        possible_progenies = set(tuple((x,)) for x in tp1_call)
    if -1 in tp1_call or -1 in tp2_call or -1 in tpp_call:
        tag = "Missing"
    else:
        tag = "Correct" if tpp_call in possible_progenies else "Error"
    return tag


def in_region(rname, rstart, target_chr, target_start, target_end):
    """
    Quick check if a point is within the target region.
    """
    return (rname == target_chr) and (target_start <= rstart <= target_end)


def alts(args):
    """
    %prog alts HD

    Build alternative loci based on simulation data.
    """
    import pysam
    from more_itertools import pairwise
    from jcvi.utils.grouper import Grouper

    p = OptionParser(alts.__doc__)
    p.set_outfile(outfile="TREDs.alts.csv")
    opts, args = p.parse_args(args)

    if len(args) < 1:
        sys.exit(not p.print_help())

    treds = args
    repo = TREDsRepo()
    if "all" in treds:
        treds = repo.names

    pad_left, pad_right = 1000, 10000
    READLEN = 150
    fw = must_open(opts.outfile, "w")
    print("TRED,alts,alts.hg19", file=fw)  # Header
    for tred in treds:
        ref_regions = []

        # Simulate a depth 1000 BAM with 300 repeats
        for ref in ("hg38", "hg19"):

            # This is the region that involves the TRED locus
            repo = TREDsRepo(ref=ref)
            t = repo[tred]
            chr, start, end = t.chr, t.repeat_start, t.repeat_end
            start -= pad_left
            end += pad_right

            tred_ref = "{}_{}".format(tred, ref)
            if not op.isdir(tred_ref):
                simulate(
                    [
                        tred_ref,
                        "300",
                        "300",
                        "--depth=1000",
                        "--ref={}".format(ref),
                        "--tred={}".format(tred),
                    ]
                )
            bamfile = op.join(tred_ref, "300.bam")

            # Parse the BAM file, retrieve all regions
            bamfile = pysam.AlignmentFile(bamfile, "rb")
            nreads = altreads = 0
            alt_points = set()
            for read in bamfile.fetch():
                fname, fstart = (
                    bamfile.getrname(read.reference_id),
                    read.reference_start,
                )
                rname, rstart = (
                    bamfile.getrname(read.next_reference_id),
                    read.next_reference_start,
                )
                f_in_region = in_region(fname, fstart, chr, start, end)
                r_in_region = in_region(rname, rstart, chr, start, end)
                if (not f_in_region) and r_in_region:
                    alt_points.add((fname, fstart))
                    altreads += 1
                if (not r_in_region) and f_in_region:
                    alt_points.add((rname, rstart))
                    altreads += 1
                nreads += 1

            logging.debug(
                "A total of {} reads ({} alts) processed".format(nreads, altreads)
            )
            alt_points = natsorted(alt_points)

            # Chain these points together into regions
            g = Grouper()
            for a in alt_points:
                g.join(a)
            for a, b in pairwise(alt_points):
                achr, apos = a
                bchr, bpos = b
                if achr != bchr:
                    continue
                if (bpos - apos) > READLEN:
                    continue
                g.join(a, b)

            # All regions that contain ALT
            alt_sum = 0
            regions = []
            for c in g:
                chr_min, pos_min = min(c)
                chr_max, pos_max = max(c)
                assert chr_min, chr_max
                pos_min -= READLEN
                pos_max += READLEN
                regions.append((chr_min, pos_min, pos_max))
                alt_sum += pos_max - pos_min

            regions = "|".join(
                [
                    "{}:{}-{}".format(c, start, end)
                    for c, start, end in natsorted(regions)
                ]
            )
            ref_regions.append(regions)

        line = ",".join([tred] + ref_regions)
        print(line, file=sys.stderr)
        print(line, file=fw)
        logging.debug("Alternative region sum: {} bp".format(alt_sum))

    fw.close()


def depth(args):
    """
    %prog depth DP.tsv

    Plot read depths across all TREDs.
    """
    import seaborn as sns

    p = OptionParser(depth.__doc__)
    opts, args, iopts = p.set_image_options(args, figsize="14x14")

    if len(args) != 1:
        sys.exit(not p.print_help())

    (tsvfile,) = args
    fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(
        ncols=2, nrows=2, figsize=(iopts.w, iopts.h)
    )
    plt.tight_layout(pad=6)

    data = pd.read_csv(tsvfile, sep="\t", low_memory=False)

    ids, treds = read_treds()
    for (dp, ax, title) in zip(
        ("FDP", "PDP", "RDP", "PEDP"),
        (ax1, ax2, ax3, ax4),
        ("Spanning reads", "Partial reads", "Repeat-only reads", "Paired-end reads"),
    ):
        logging.debug("Build {}".format(title))
        # Construct related data structure
        xd = []  # (tred, dp)
        mdp = []  # (tred, median_dp)
        for tred, motif in zip(treds["abbreviation"], treds["motif"]):
            if tred in ignore:
                logging.debug("Ignore {}".format(tred))
                continue
            if len(motif) > 4:
                if "/" in motif:  # CTG/CAG
                    motif = motif.split("/")[0]
                else:
                    motif = motif[:4] + ".."
            xtred = "{} {}".format(tred, motif)
            md = [x for x in data[tred + "." + dp] if x >= 0]
            subsample = 10000 if dp == "RDP" else 1000
            md = sample(md, subsample)
            pmd = [x for x in md if x > 0]
            median = np.median(pmd) if pmd else 0
            mdp.append((xtred, median))
            for d in md:
                xd.append((xtred, d))

        # Determine order
        mdp.sort(key=lambda x: x[1])
        order, mdp = zip(*mdp)

        # OK, now plot
        xt, xd = zip(*xd)
        sns.boxplot(xt, xd, ax=ax, order=order, fliersize=2)
        xticklabels = ax.get_xticklabels()
        ax.set_xticklabels(xticklabels, rotation=45, ha="right")
        ax.set_title("Number of {} per locus".format(title), size=18)
        ylim = 30 if dp == "RDP" else 100
        ax.set_ylim(0, ylim)

        yticklabels = [int(x) for x in ax.get_yticks()]
        ax.set_yticklabels(yticklabels, family="Helvetica", size=14)

    root = fig.add_axes([0, 0, 1, 1])
    pad = 0.04
    panel_labels(
        root,
        (
            (pad, 1 - pad, "A"),
            (1 / 2.0 + pad / 2, 1 - pad, "B"),
            (pad, 0.5 - pad / 2, "C"),
            (1 / 2.0 + pad / 2, 0.5 - pad / 2, "D"),
        ),
    )
    normalize_axes(root)

    image_name = "depth." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def mendelian_errors(args):
    """
    %prog mendelian_errors STR-Mendelian-errors.csv

    Plot Mendelian errors as calculated by mendelian(). File
    `STR-Mendelian-errors.csv` looks like:

    ,Duos  - Mendelian errors,Trios - Mendelian errors
    SCA36,1.40%,0.60%
    ULD,0.30%,1.50%
    BPES,0.00%,1.80%

    One TRED disease per line, followed by duo errors and trio errors.
    """
    p = OptionParser(mendelian_errors.__doc__)
    opts, args, iopts = p.set_image_options(args, figsize="6x6")

    if len(args) != 1:
        sys.exit(not p.print_help())

    (csvfile,) = args
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])

    ymin = -0.2
    df = pd.read_csv(csvfile)
    data = []
    for i, d in df.iterrows():
        if d["TRED"].split()[0] in ignore:
            logging.debug("Ignore {}".format(d["TRED"]))
            continue
        data.append(d)
    treds, duos, trios = zip(*data)
    ntreds = len(treds)
    ticks = range(ntreds)
    treds = [x.split()[0] for x in treds]
    duos = [float(x.rstrip("%")) for x in duos]
    trios = [float(x.rstrip("%")) for x in trios]

    for tick, duo, trio in zip(ticks, duos, trios):
        m = max(duo, trio)
        ax.plot([tick, tick], [ymin, m], "-", lw=2, color="lightslategray")

    (duos,) = ax.plot(duos, "o", mfc="w", mec="g")
    (trios,) = ax.plot(trios, "o", mfc="w", mec="b")
    ax.set_title("Mendelian errors based on trios and duos in HLI samples")
    nduos = "Mendelian errors in 362 duos"
    ntrios = "Mendelian errors in 339 trios"
    ax.legend([trios, duos], [ntrios, nduos], loc="best")

    ax.set_xticks(ticks)
    ax.set_xticklabels(treds, rotation=45, ha="right", size=8)
    yticklabels = [int(x) for x in ax.get_yticks()]
    ax.set_yticklabels(yticklabels, family="Helvetica")
    ax.set_ylabel(r"Mendelian errors (\%)")
    ax.set_ylim(ymin, 20)

    normalize_axes(root)

    image_name = "mendelian_errors." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def extract_trios(family):
    """
    Identify all trios/duos inside a family, where a family contains dictionary
    of relationship: individual, for example:
      {
        "ChildSelf": "176531498",
        "DzTwin": "176531497",
        "Parent": "176449143"
      }
    """
    self_key = ["ChildSelf"]
    keys = family.keys()
    spouse_key = [x for x in keys if ("spouse" in x.lower())]
    assert len(spouse_key) <= 1
    parent_keys = [
        x for x in keys if ("parent" in x.lower()) and ("grand" not in x.lower())
    ]
    sib_keys = [
        x for x in keys if ("sibling" in x.lower()) or ("twin" in x.lower())
    ] + self_key
    child_keys = [
        x
        for x in keys
        if ("child" in x.lower())
        and ("grand" not in x.lower())
        and ("self" not in x.lower())
    ]

    for sk in sib_keys:
        yield TrioOrDuo(parent_keys, [sk], family)
    for ck in child_keys:
        yield TrioOrDuo(self_key + spouse_key, [ck], family)


def read_tred_tsv(tsvfile):
    """
    Read the TRED table into a dataframe.
    """
    df = pd.read_csv(tsvfile, sep="\t", index_col=0, dtype={"SampleKey": str})
    return df


def mendelian(args):
    """
    %prog mendelian trios_candidate.json hli.20170424.tred.tsv

    Calculate Mendelian errors based on trios and duos.
    """
    p = OptionParser(mendelian.__doc__)
    p.add_option("--tolerance", default=0, type="int", help="Tolernace for differences")
    p.set_verbose()
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    triosjson, tredtsv = args
    verbose = opts.verbose
    tolerance = opts.tolerance

    js = json.load(open(triosjson))
    allterms = set()
    duos = set()
    trios = set()
    for v in js:
        allterms |= set(v.keys())
        for trio_or_duo in extract_trios(v):
            assert len(trio_or_duo) in (2, 3)
            if len(trio_or_duo) == 2:
                duos.add(trio_or_duo)
            else:
                trios.add(trio_or_duo)
    # print "\n".join(allterms)
    print("A total of {} families imported".format(len(js)))

    # Read in all data
    df = read_tred_tsv(tredtsv)

    ids, treds = read_treds()
    table = {}
    for tred, inheritance in zip(treds["abbreviation"], treds["inheritance"]):
        x_linked = inheritance[0] == "X"  # X-linked
        name = tred
        if x_linked:
            name += " (X-linked)"
        print("[TRED] {}".format(name))

        n_total = len(duos)
        n_error = 0
        for duo in duos:
            n_error += duo.check_mendelian(
                df, tred, tolerance=tolerance, x_linked=x_linked, verbose=verbose
            )
        tag = "Duos  - Mendelian errors"
        print("{}: {}".format(tag, percentage(n_error, n_total)))
        duo_error = percentage(n_error, n_total, mode=2)
        table[(name, tag)] = "{0:.1f}%".format(duo_error)

        n_total = len(trios)
        n_error = 0
        for trio in trios:
            n_error += trio.check_mendelian(
                df, tred, tolerance=tolerance, x_linked=x_linked, verbose=verbose
            )
        tag = "Trios - Mendelian errors"
        print("{}: {}".format(tag, percentage(n_error, n_total)))
        trio_error = percentage(n_error, n_total, mode=2)
        table[(name, tag)] = "{0:.1f}%".format(trio_error)

    # Summarize
    print(tabulate(table))


def make_STR_bed(filename="STR.bed", pad=0, treds=None):
    tredsfile = datafile("TREDs.meta.csv")
    tf = pd.read_csv(tredsfile)

    tds = list(tf["abbreviation"])
    regions = list(tf["repeat_location"])
    fw = must_open(filename, "w")
    extract_Y = False
    for td, region in zip(tds, regions):
        if treds and (td not in treds):
            continue
        c, startend = region.split(":")
        extract_Y = extract_Y or (c == "chrY")
        start, end = startend.split("-")
        start, end = int(start), int(end)
        print("\t".join(str(x) for x in (c, start - pad, end + pad, td)), file=fw)

    if not extract_Y:
        return filename

    UNIQY = datafile("chrY.hg38.unique_ccn.gc")
    fp = open(UNIQY)
    nregions = 0
    for i, row in enumerate(fp):
        # Some regions still have mapped reads, exclude a few
        if i in (1, 4, 6, 7, 10, 11, 13, 16, 18, 19):
            continue
        if nregions >= 5:
            break
        c, start, end, gc = row.split()
        start, end = int(start), int(end)
        print(
            "\t".join(
                str(x)
                for x in (
                    c,
                    start - pad,
                    end + pad,
                    "chrY.unique_ccn.{}".format(nregions),
                )
            ),
            file=fw,
        )
        nregions += 1

    fw.close()
    return filename


def mini(args):
    """
    %prog mini bamfile minibamfile

    Prepare mini-BAMs that contain only the STR loci.
    """
    p = OptionParser(mini.__doc__)
    p.add_option(
        "--pad", default=20000, type="int", help="Add padding to the STR reigons"
    )
    p.add_option(
        "--treds", default=None, help="Extract specific treds, use comma to separate"
    )
    p.set_outfile()
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    bamfile, minibam = args
    treds = opts.treds.split(",") if opts.treds else None
    pad = opts.pad
    bedfile = make_STR_bed(pad=pad, treds=treds)

    get_minibam_bed(bamfile, bedfile, minibam)
    logging.debug("Mini-BAM written to `{}`".format(minibam))


def parse_log(logfile):
    fp = open(logfile)
    likelihood = {}
    for row in fp:
        if row.startswith("DEBUG:IntegratedCaller:***"):
            atoms = row.split()
            i = int(atoms[1].strip("(,"))
            j = int(atoms[2].strip(")"))
            lnL = float(atoms[-1])
            likelihood[(i, j)] = lnL
        if row.startswith("DEBUG:IntegratedCaller:CI(h1)"):
            CI_h1 = [int(x.strip()) for x in row.split("=")[1].split("-")]
        if row.startswith("DEBUG:IntegratedCaller:CI(h2)"):
            CI_h2 = [int(x.strip()) for x in row.split("=")[1].split("-")]
        if row.startswith("DEBUG:IntegratedCaller:ML estimate:"):
            MLE = row.split(":")[3].split("=")[1].split()[:2]
            MLE = [int(x.strip("[],")) for x in MLE]

    return likelihood, CI_h1, CI_h2, MLE


def likelihood(args):
    """
    %prog likelihood

    Plot likelihood surface. Look for two files in the current folder:
    - 100_100.log, haploid model
    - 100_20.log, diploid model
    """
    p = OptionParser(likelihood.__doc__)
    opts, args, iopts = p.set_image_options(
        args, figsize="10x5", style="white", cmap="coolwarm"
    )

    if len(args) != 0:
        sys.exit(not p.print_help())

    fig, (ax1, ax2) = plt.subplots(ncols=2, nrows=1, figsize=(iopts.w, iopts.h))
    plt.tight_layout(pad=4)

    # Haploid model
    LL, CI_h1, CI_h2, MLE = parse_log("100_100.log")
    data = []
    for k, v in LL.items():
        data.append((k[0], v))
    data.sort()
    x, y = zip(*data)
    x = np.array(x)
    (curve,) = ax1.plot(x, y, "-", color=lsg, lw=2)
    ax1.set_title("Simulated haploid ($h^{truth}=100$)")

    h_hat, max_LL = max(data, key=lambda x: x[-1])
    _, min_LL = min(data, key=lambda x: x[-1])
    ymin, ymax = ax1.get_ylim()
    ax1.set_ylim([ymin, ymax + 30])

    LL_label = "log(Likelihood)"
    ax1.plot([h_hat, h_hat], [ymin, max_LL], ":", color=lsg, lw=2)
    ax1.text(h_hat, max_LL + 10, r"$\hat{h}=93$", color=lsg)
    ax1.set_xlabel(r"$h$")
    ax1.set_ylabel(LL_label)

    a, b = CI_h1
    ci = ax1.fill_between(
        x, [ymin] * len(x), y, where=(x >= a) & (x <= b), color=lsg, alpha=0.5
    )
    ax1.legend([curve, ci], ["Likelihood curve", r"95$\%$ CI"], loc="best")

    # Diploid model
    LL, CI_h1, CI_h2, MLE = parse_log("100_20.log")
    _, min_LL = min(data, key=lambda x: x[-1])
    data = np.ones((301, 301)) * min_LL
    for k, v in LL.items():
        a, b = k
        data[a, b] = v
        data[b, a] = v

    data = mask_upper_triangle(data)
    ax_imshow(ax2, data, opts.cmap, LL_label, 20, 104)

    root = fig.add_axes([0, 0, 1, 1])
    pad = 0.04
    panel_labels(root, ((pad / 2, 1 - pad, "A"), (1 / 2.0, 1 - pad, "B")))
    normalize_axes(root)

    image_name = "likelihood." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def mask_upper_triangle(data):
    mask = np.zeros_like(data)
    mask[np.triu_indices_from(mask)] = True
    data = np.ma.array(data, mask=mask)
    return data


def ax_plot(ax, P_h, h_hat, CI_h, xlabel, ylabel, ticks=True):
    max_P = max(P_h.values())
    a, b = CI_h

    ax.plot([h_hat, h_hat], [0, max_P], ":", color=lsg, lw=2)
    ax.set_xlabel(r"$%s$" % xlabel)
    ax.set_ylabel(ylabel)

    data = []
    for k, v in sorted(P_h.items()):
        data.append((int(k), v))
    data.sort()
    x, y = zip(*data)
    x = np.array(x)
    ax.plot(x, y, "-", color=lsg, lw=2)
    title = "Marginal distribution for $%s$" % xlabel
    ax.set_title(title)
    if not ticks:
        ax.set_yticks([])

    if a == b:
        ax.plot([h_hat, h_hat], [0, max_P], "-", color=lsg, lw=2)
    else:
        ax.fill_between(
            x, [0] * len(x), y, where=(x >= a) & (x <= b), color=lsg, alpha=0.5
        )
    ax.set_xlim(0, 300)

    ymin, ymax = ax.get_ylim()
    if h_hat < 150:
        ax.text(
            h_hat + 20,
            ymax * 4.0 / 5,
            r"$\hat{%s}=%d$" % (xlabel, h_hat),
            color=lsg,
            va="center",
        )
        ax.text(
            h_hat + 20,
            ymax * 3.0 / 5,
            r"95$\%$ CI" + r"$=%s-%s$" % (a, b),
            color=lsg,
            va="center",
        )
    else:
        ax.text(
            h_hat - 30,
            ymax * 4.0 / 5,
            r"$\hat{%s}=%d$" % (xlabel, h_hat),
            color=lsg,
            ha="right",
            va="center",
        )
        ax.text(
            h_hat - 30,
            ymax * 3.0 / 5,
            r"95$\%$ CI" + r"$=%s-%s$" % (a, b),
            color=lsg,
            ha="right",
            va="center",
        )

    ymin, ymax = ax.get_ylim()
    ax.set_ylim(ymin, ymax * 1.05)


def ax_imshow(
    ax,
    P_h1h2,
    cmap,
    label,
    h1_hat,
    h2_hat,
    h1_truth,
    h2_truth,
    r=4,
    draw_circle=True,
    ticks=True,
):
    im = ax.imshow(P_h1h2, cmap=cmap, origin="lower")

    from mpl_toolkits.axes_grid1 import make_axes_locatable

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cb = plt.colorbar(im, cax)
    cb.set_label(label)
    if not ticks:
        cb.set_ticks([])

    if draw_circle:
        circle = plt.Circle((h1_hat, h2_hat), r, ec="w", fill=False)
        ax.add_artist(circle)

    annotation = r"$\hat{h_1}=%d, \hat{h_2}=%d$" % (h1_hat, h2_hat)
    ax.text(200, 100, annotation, color=lsg, ha="center", va="center")

    ax.set_xlabel(r"$h_1$")
    ax.set_ylabel(r"$h_2$")
    title = "Simulated diploid ($h_{1}^{truth}=%d, h_{2}^{truth}=%d$)" % (
        h1_truth,
        h2_truth,
    )
    ax.set_title(title)


def likelihood2(args):
    """
    %prog likelihood2 100_20.json

    Plot the likelihood surface and marginal distributions.
    """
    from matplotlib import gridspec

    p = OptionParser(likelihood2.__doc__)
    opts, args, iopts = p.set_image_options(
        args, figsize="10x5", style="white", cmap="coolwarm"
    )

    if len(args) != 1:
        sys.exit(not p.print_help())

    (jsonfile,) = args
    fig = plt.figure(figsize=(iopts.w, iopts.h))
    gs = gridspec.GridSpec(2, 2)
    ax1 = fig.add_subplot(gs[:, 0])
    ax2 = fig.add_subplot(gs[0, 1])
    ax3 = fig.add_subplot(gs[1, 1])
    plt.tight_layout(pad=3)
    pf = plot_panel(jsonfile, ax1, ax2, ax3, opts.cmap)

    root = fig.add_axes([0, 0, 1, 1])
    normalize_axes(root)

    image_name = "likelihood2.{}.".format(pf) + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def likelihood3(args):
    """
    %prog likelihood3 140_20.json 140_70.json

    Plot the likelihood surface and marginal distributions for two settings.
    """
    from matplotlib import gridspec

    p = OptionParser(likelihood3.__doc__)
    opts, args, iopts = p.set_image_options(
        args, figsize="10x10", style="white", cmap="coolwarm"
    )
    if len(args) != 2:
        sys.exit(not p.print_help())

    jsonfile1, jsonfile2 = args
    fig = plt.figure(figsize=(iopts.w, iopts.h))
    gs = gridspec.GridSpec(9, 2)
    ax1 = fig.add_subplot(gs[:4, 0])
    ax2 = fig.add_subplot(gs[:2, 1])
    ax3 = fig.add_subplot(gs[2:4, 1])
    ax4 = fig.add_subplot(gs[5:, 0])
    ax5 = fig.add_subplot(gs[5:7, 1])
    ax6 = fig.add_subplot(gs[7:, 1])
    plt.tight_layout(pad=2)

    plot_panel(jsonfile1, ax1, ax2, ax3, opts.cmap)
    plot_panel(jsonfile2, ax4, ax5, ax6, opts.cmap)

    root = fig.add_axes([0, 0, 1, 1])
    pad = 0.02
    panel_labels(root, ((pad, 1 - pad, "A"), (pad, 4.0 / 9, "B")))
    normalize_axes(root)

    image_name = "likelihood3." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def plot_panel(jsonfile, ax1, ax2, ax3, cmap, tred="HD"):
    j = json.load(open(jsonfile))
    calls = j["tredCalls"]
    P_h1h2 = calls[tred + ".P_h1h2"]
    data = np.zeros((301, 301))
    for k, v in P_h1h2.items():
        a, b = k.split(",")
        a, b = int(a), int(b)
        data[a, b] = v
        data[b, a] = v

    label = "Probability density"
    data = mask_upper_triangle(data)
    h1_hat, h2_hat = calls[tred + ".1"], calls[tred + ".2"]
    pf = op.basename(jsonfile).split(".")[0]
    h1_truth, h2_truth = sorted([int(x) for x in pf.split("_")])
    ax_imshow(
        ax1,
        data,
        cmap,
        label,
        h1_hat,
        h2_hat,
        h1_truth,
        h2_truth,
        draw_circle=False,
        ticks=False,
    )

    CI = calls[tred + ".CI"]
    CI_h1, CI_h2 = CI.split("|")
    CI_h1 = [int(x) for x in CI_h1.split("-")]
    CI_h2 = [int(x) for x in CI_h2.split("-")]
    P_h1 = calls[tred + ".P_h1"]
    P_h2 = calls[tred + ".P_h2"]

    ax_plot(ax2, P_h1, h1_hat, CI_h1, "h_1", label, ticks=False)
    ax_plot(ax3, P_h2, h2_hat, CI_h2, "h_2", label, ticks=False)

    return pf


def diagram(args):
    """
    %prog diagram

    Plot the predictive power of various evidences.
    """
    p = OptionParser(diagram.__doc__)
    opts, args, iopts = p.set_image_options(args, figsize="8x4", format="png")

    if len(args) != 0:
        sys.exit(not p.print_help())

    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])

    # Gauge on top, this is log-scale
    yy = 0.7
    yinterval = 0.1
    height = 0.05
    yp = yy - yinterval - height
    canvas = 0.95
    xstart = 0.025
    convert = lambda x: xstart + x * canvas / 600
    # Symbols
    root.text(
        0.5, 0.9, r"$L$: Read length, $F$: Flank size, $V$: Pair distance", ha="center"
    )
    root.text(0.5, 0.85, r"ex. $L=150bp, F=9bp, V=500bp$", ha="center")
    root.text(
        xstart + canvas,
        yy - height,
        "STR repeat length",
        ha="center",
        color=lsg,
        size=10,
    )

    # Mark the key events
    pad = 0.02
    arrowlen = canvas * 1.05
    arrowprops = dict(
        length_includes_head=True,
        width=0.01,
        fc=lsg,
        lw=0,
        head_length=arrowlen * 0.12,
        head_width=0.04,
    )
    p = FancyArrow(xstart, yy, arrowlen, 0, shape="right", **arrowprops)
    root.add_patch(p)

    ppad = 30
    keyevents = (
        (0, 0, -1, r"$0$"),
        (150 - 18, 150 - 18 - ppad, 0, r"$L - 2F$"),
        (150 - 9, 150 - 9, 1, r"$L - F$"),
        (150, 150 + ppad, 2, r"$L$"),
        (500 - 9, 500 - 9, 3, r"$V - F$"),
    )
    for event, pos, i, label in keyevents:
        _event = convert(event)
        _pos = convert(pos)
        root.plot((_event, _event), (yy - height / 4, yy + height / 4), "-", color="k")
        root.text(_pos, yy + pad, label, rotation=45, va="bottom", size=8)
        if i < 0:
            continue
        ystart = yp - i * yinterval
        root.plot((_event, _event), (ystart, yy - height / 4), ":", color=lsg)

    # Range on bottom. These are simple 4 rectangles, with the range indicating
    # the predictive range.
    CLOSED, OPEN = range(2)
    ranges = (
        (0, 150 - 18, CLOSED, "Spanning reads"),
        (9, 150 - 9, OPEN, "Partial reads"),
        (150, 500 - 9, CLOSED, "Repeat reads"),
        (0, 500 - 9, CLOSED, "Paired-end reads"),
    )
    for start, end, starttag, label in ranges:
        _start = convert(start)
        _end = convert(end)
        data = (
            [[0.0, 1.0], [0.0, 1.0]] if starttag == OPEN else [[1.0, 0.0], [1.0, 0.0]]
        )
        root.imshow(
            data,
            interpolation="bicubic",
            cmap=plt.cm.Greens,
            extent=[_start, _end, yp, yp + height],
        )
        root.text(_end + pad, yp + height / 2, label, va="center")
        yp -= yinterval

    normalize_axes(root)

    image_name = "diagram." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def plot_allelefreq(ax, df, locus, color="lightslategray"):
    tred = df.ix[locus]
    cnt = af_to_counts(tred["allele_freq"])

    cntx, cnty = zip(*cnt.items())

    motif = tred["motif"]
    cutoff_prerisk = tred["cutoff_prerisk"]
    cutoff_risk = tred["cutoff_risk"]
    npredisease = sum(v for (k, v) in cnt.items() if cutoff_prerisk <= k < cutoff_risk)
    npatients = sum(v for (k, v) in cnt.items() if k >= cutoff_risk)

    ax.bar(cntx, cnty, fc=color)

    ymin, ymax = ax.get_ylim()
    xmax = (cutoff_risk / 10 + 1) * 10 if cutoff_risk > 50 else 50
    pad = xmax * 0.03
    if cutoff_prerisk < cutoff_risk and npredisease:
        ax.axvline(x=cutoff_prerisk, color="k", lw=2)
        ax.text(
            cutoff_prerisk + pad,
            0.5 * ymax,
            r"Pre-disease ($\geq${}$\times${}) - {} alleles".format(
                cutoff_prerisk, motif, npredisease
            ),
            rotation=90,
            color="k",
            ha="center",
            va="center",
        )
    ax.axvline(x=cutoff_risk, color="r", lw=2)

    if locus == "AR":
        npatients = sum(v for (k, v) in cnt.items() if k <= cutoff_risk)
        ax.text(
            cutoff_risk - pad,
            0.5 * ymax,
            r"Disease ($\leq${}$\times${}) - {} alleles".format(
                cutoff_risk, motif, npatients
            ),
            rotation=90,
            color="r",
            ha="center",
            va="center",
        )
    else:
        ax.text(
            cutoff_risk + pad,
            0.5 * ymax,
            r"Disease ($\geq${}$\times${}) - {} alleles".format(
                cutoff_risk, motif, npatients
            ),
            rotation=90,
            color="r",
            ha="center",
            va="center",
        )

    x = []  # All allelels
    for k, v in cnt.items():
        x.extend([k] * v)

    ax.set_xlabel("Number of repeat units")
    ax.set_ylabel("Number of alleles")
    ax.set_xlim(0, xmax)
    ax.set_title(r"{} ({})".format(locus, tred["title"], motif))
    set_helvetica_axis(ax)


def allelefreqall(args):
    """
    %prog allelefreqall HN_Platinum_Gold.20180525.tsv.report.txt

    Plot all 30 STR allele frequencies.
    """
    p = OptionParser(allelefreqall.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (reportfile,) = args
    treds, df = read_treds(reportfile)
    # Prepare 5 pages, each page with 6 distributions
    treds = sorted(treds)
    count = 6
    pdfs = []
    for page in range(len(treds) / count + 1):
        start = page * count
        page_treds = treds[start : start + count]
        if not page_treds:
            break
        allelefreq(
            [
                ",".join(page_treds),
                "--usereport",
                reportfile,
                "--nopanels",
                "--figsize",
                "12x16",
            ]
        )
        outpdf = "allelefreq.{}.pdf".format(page)
        sh("mv allelefreq.pdf {}".format(outpdf))
        pdfs.append(outpdf)

    from jcvi.formats.pdf import cat

    pf = op.basename(reportfile).split(".")[0]
    finalpdf = pf + ".allelefreq.pdf"
    logging.debug("Merging pdfs into `{}`".format(finalpdf))
    cat(pdfs + ["-o", finalpdf, "--cleanup"])


def allelefreq(args):
    """
    %prog allelefreq HD,DM1,SCA1,SCA17,FXTAS,FRAXE

    Plot the allele frequencies of some STRs.
    """
    p = OptionParser(allelefreq.__doc__)
    p.add_option(
        "--nopanels",
        default=False,
        action="store_true",
        help="No panel labels A, B, ...",
    )
    p.add_option("--usereport", help="Use allele frequency in report file")
    opts, args, iopts = p.set_image_options(args, figsize="9x13")

    if len(args) != 1:
        sys.exit(not p.print_help())

    (loci,) = args
    fig, ((ax1, ax2), (ax3, ax4), (ax5, ax6)) = plt.subplots(
        ncols=2, nrows=3, figsize=(iopts.w, iopts.h)
    )
    plt.tight_layout(pad=4)
    if opts.usereport:
        treds, df = read_treds(tredsfile=opts.usereport)
    else:
        treds, df = read_treds()

    df = df.set_index(["abbreviation"])

    axes = (ax1, ax2, ax3, ax4, ax5, ax6)
    loci = loci.split(",")
    for ax, locus in zip(axes, loci):
        plot_allelefreq(ax, df, locus)

    # Delete unused axes
    for ax in axes[len(loci) :]:
        ax.set_axis_off()

    root = fig.add_axes([0, 0, 1, 1])
    pad = 0.03
    if not opts.nopanels:
        panel_labels(
            root,
            (
                (pad / 2, 1 - pad, "A"),
                (0.5 + pad, 1 - pad, "B"),
                (pad / 2, 2 / 3.0 - pad / 2, "C"),
                (0.5 + pad, 2 / 3.0 - pad / 2, "D"),
                (pad / 2, 1 / 3.0, "E"),
                (0.5 + pad, 1 / 3.0, "F"),
            ),
        )
    normalize_axes(root)

    image_name = "allelefreq." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def make_fasta(seq, fastafile, id):
    rec = SeqRecord(Seq(seq), description="", id=id)
    fw = open(fastafile, "w")
    SeqIO.write([rec], fw, "fasta")
    fw.close()


def add_simulate_options(p):
    p.add_option("--readlen", default=150, type="int", help="Length of the read")
    p.add_option(
        "--distance",
        default=500,
        type="int",
        help="Outer distance between the two ends",
    )
    p.set_depth(depth=20)


def simulate(args):
    """
    %prog simulate run_dir 1 300

    Simulate BAMs with varying inserts with dwgsim. The above command will
    simulate between 1 to 300 CAGs in the HD region, in a directory called
    `run_dir`.
    """
    p = OptionParser(simulate.__doc__)
    p.add_option(
        "--method", choices=("wgsim", "eagle"), default="eagle", help="Read simulator"
    )
    p.add_option(
        "--ref",
        default="hg38",
        choices=("hg38", "hg19"),
        help="Reference genome version",
    )
    p.add_option("--tred", default="HD", help="TRED locus")
    add_simulate_options(p)
    opts, args = p.parse_args(args)

    if len(args) != 3:
        sys.exit(not p.print_help())

    rundir, startunits, endunits = args
    ref = opts.ref
    ref_fasta = "/mnt/ref/{}.upper.fa".format(ref)
    startunits, endunits = int(startunits), int(endunits)
    basecwd = os.getcwd()
    mkdir(rundir)
    os.chdir(rundir)
    cwd = os.getcwd()

    # TRED region (e.g. Huntington)
    pad_left, pad_right = 1000, 10000
    repo = TREDsRepo(ref=ref)
    tred = repo[opts.tred]
    chr, start, end = tred.chr, tred.repeat_start, tred.repeat_end

    logging.debug("Simulating {}".format(tred))
    fasta = Fasta(ref_fasta)
    seq_left = fasta[chr][start - pad_left : start - 1]
    seq_right = fasta[chr][end : end + pad_right]
    motif = tred.repeat

    simulate_method = wgsim if opts.method == "wgsim" else eagle
    # Write fake sequence
    for units in range(startunits, endunits + 1):
        pf = str(units)
        mkdir(pf)
        os.chdir(pf)
        seq = str(seq_left) + motif * units + str(seq_right)
        fastafile = pf + ".fasta"
        make_fasta(seq, fastafile, id=chr.upper())

        # Simulate reads on it
        simulate_method(
            [
                fastafile,
                "--depth={}".format(opts.depth),
                "--readlen={}".format(opts.readlen),
                "--distance={}".format(opts.distance),
                "--outfile={}".format(pf),
            ]
        )

        read1 = pf + ".bwa.read1.fastq"
        read2 = pf + ".bwa.read2.fastq"
        samfile, _ = align([ref_fasta, read1, read2])
        indexed_samfile = index([samfile])

        sh("mv {} ../{}.bam".format(indexed_samfile, pf))
        sh("mv {}.bai ../{}.bam.bai".format(indexed_samfile, pf))

        os.chdir(cwd)
        shutil.rmtree(pf)

    os.chdir(basecwd)


def mergebam(args):
    """
    %prog mergebam dir1 homo_outdir
    or
    %prog mergebam dir1 dir2/20.bam het_outdir

    Merge sets of BAMs to make diploid. Two modes:
    - Homozygous mode: pair-up the bams in the two folders and merge
    - Heterozygous mode: pair the bams in first folder with a particular bam
    """
    p = OptionParser(mergebam.__doc__)
    p.set_cpus()
    opts, args = p.parse_args(args)

    if len(args) not in (2, 3):
        sys.exit(not p.print_help())

    if len(args) == 2:
        idir1, outdir = args
        dir1 = [idir1] if idir1.endswith(".bam") else iglob(idir1, "*.bam")
        logging.debug("Homozygous mode")
        dir2 = [""] * len(dir1)
    elif len(args) == 3:
        idir1, idir2, outdir = args
        dir1 = [idir1] if idir1.endswith(".bam") else iglob(idir1, "*.bam")
        dir2 = [idir2] if idir2.endswith(".bam") else iglob(idir2, "*.bam")
        assert len(dir2) == 1, "Second pile must contain a single bam"
        dir2 = [idir2] * len(dir1)

    assert len(dir1) == len(dir2), "Two piles must contain same number of bams"
    cmd = "samtools merge {} {} {} && samtools index {}"
    cmds = []
    mkdir(outdir)
    for a, b in zip(dir1, dir2):
        ia = op.basename(a).split(".")[0]
        ib = op.basename(b).split(".")[0] if b else ia
        outfile = op.join(outdir, "{}_{}.bam".format(ia, ib))
        cmds.append(cmd.format(outfile, a, b, outfile))

    p = Parallel(cmds, cpus=opts.cpus)
    p.run()


def batchlobstr(args):
    """
    %prog batchlobstr bamlist

    Run lobSTR on a list of BAMs. The corresponding batch command for TREDPARSE:
    $ tred.py bamlist --haploid chr4 --workdir tredparse_results
    """
    p = OptionParser(batchlobstr.__doc__)
    p.add_option(
        "--haploid", default="chrY,chrM", help="Use haploid model for these chromosomes"
    )
    p.set_cpus()
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (bamlist,) = args
    cmd = "python -m jcvi.variation.str lobstr TREDs"
    cmd += " --input_bam_path {}"
    cmd += " --haploid {}".format(opts.haploid)
    cmd += " --simulation"
    cmds = [cmd.format(x.strip()) for x in open(bamlist).readlines()]
    p = Parallel(cmds, cpus=opts.cpus)
    p.run()


def compilevcf(args):
    """
    %prog compilevcf dir

    Compile vcf outputs into lists.
    """
    from jcvi.variation.str import LobSTRvcf

    p = OptionParser(compilevcf.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (folder,) = args
    vcf_files = iglob(folder, "*.vcf,*.vcf.gz")
    for vcf_file in vcf_files:
        try:
            p = LobSTRvcf(columnidsfile=None)
            p.parse(vcf_file, filtered=False)
            res = p.items()
            if res:
                k, v = res[0]
                res = v.replace(",", "/")
            else:
                res = "-1/-1"
            num = op.basename(vcf_file).split(".")[0]
            print(num, res)
        except (TypeError, AttributeError) as e:
            p = TREDPARSEvcf(vcf_file)
            continue


def evidences(args):
    """
    %prog evidences

    Plot distribution of evidences against two factors:
    - Sample mean coverage
    - Longer allele
    """
    p = OptionParser(evidences.__doc__)
    p.add_option(
        "--csv", default="hli.20170328.tred.tsv", help="TRED csv output to plot"
    )
    opts, args, iopts = p.set_image_options(args, format="pdf")

    if len(args) != 0:
        sys.exit(not p.print_help())

    format = iopts.format

    # Extract sample coverage first
    df = pd.read_csv(
        "qc-export-MeanCoverage.csv",
        header=None,
        names=["Samplekey", "MeanCoverage"],
        index_col=0,
    )

    # Find coverage for HD
    xf = pd.read_csv(opts.csv, sep="\t", index_col=0)
    dp = {}
    tred = "HD"
    for sk, row in xf.iterrows():
        sk = str(sk)
        a1 = row[tred + ".1"]
        a2 = row[tred + ".2"]
        fdp = row[tred + ".FDP"]
        pdp = row[tred + ".PDP"]
        pedp = row[tred + ".PEDP"]
        dp[sk] = (a1, a2, fdp, pdp, pedp)

    # Build a consolidated dataframe
    ef = pd.DataFrame.from_dict(dp, orient="index")
    ef.columns = [
        tred + ".1",
        tred + ".2",
        tred + ".FDP",
        tred + ".PDP",
        tred + ".PEDP",
    ]
    ef.index.name = "SampleKey"
    mf = df.merge(ef, how="right", left_index=True, right_index=True)

    # Plot a bunch of figures
    outdir = "output"
    mkdir(outdir)
    xlim = ylim = (0, 100)
    draw_jointplot(
        outdir + "/A",
        "MeanCoverage",
        "HD.FDP",
        data=mf,
        xlim=xlim,
        ylim=ylim,
        format=format,
    )
    draw_jointplot(
        outdir + "/B",
        "MeanCoverage",
        "HD.PDP",
        data=mf,
        color="g",
        xlim=xlim,
        ylim=ylim,
        format=format,
    )
    draw_jointplot(
        outdir + "/C",
        "MeanCoverage",
        "HD.PEDP",
        data=mf,
        color="m",
        xlim=xlim,
        ylim=ylim,
        format=format,
    )

    xlim = (0, 50)
    draw_jointplot(
        outdir + "/D", "HD.2", "HD.FDP", data=mf, xlim=xlim, ylim=ylim, format=format
    )
    draw_jointplot(
        outdir + "/E",
        "HD.2",
        "HD.PDP",
        data=mf,
        color="g",
        xlim=xlim,
        ylim=ylim,
        format=format,
    )
    draw_jointplot(
        outdir + "/F",
        "HD.2",
        "HD.PEDP",
        data=mf,
        color="m",
        xlim=xlim,
        ylim=ylim,
        format=format,
    )


def draw_jointplot(
    figname, x, y, data=None, kind="reg", color=None, xlim=None, ylim=None, format="pdf"
):
    """
    Wraps around sns.jointplot
    """
    import seaborn as sns

    sns.set_context("talk")
    plt.clf()

    register = {
        "MeanCoverage": "Sample Mean Coverage",
        "HD.FDP": "Depth of full spanning reads",
        "HD.PDP": "Depth of partial spanning reads",
        "HD.PEDP": "Depth of paired-end reads",
        "HD.2": "Repeat size of the longer allele",
    }

    g = sns.jointplot(x, y, data=data, kind=kind, color=color, xlim=xlim, ylim=ylim)
    g.ax_joint.set_xlabel(register.get(x, x))
    g.ax_joint.set_ylabel(register.get(y, y))
    savefig(figname + "." + format, cleanup=False)


def long_allele(s, default=19, exclude=None):
    if "_" in s:
        a, b = s.split("_")
    elif "/" in s:
        a, b = s.split("/")
    else:
        raise Exception("Don't know how to split string {}".format(s))

    res = [int(a), int(b)]
    if exclude and exclude in res:
        res.remove(exclude)
    res = max(res)
    return default if res < 0 else res


def get_lo_hi_from_CI(s, exclude=None):
    """
    Parse the confidence interval from CI.

    >>> get_lo_hi_from_CI("20-20/40-60")
    (40, 60)
    """
    a, b = s.split("|")
    ai, aj = a.split("-")
    bi, bj = b.split("-")

    los = [int(ai), int(bi)]
    his = [int(aj), int(bj)]
    if exclude and exclude in los:
        los.remove(exclude)
    if exclude and exclude in his:
        his.remove(exclude)
    return max(los), max(his)


def parse_results(datafile, exclude=None):
    fp = open(datafile)
    data = []
    for row in fp:
        atoms = row.split()
        truth, call = atoms[:2]
        t = long_allele(truth, exclude=exclude)
        c = long_allele(call, exclude=exclude)
        if len(atoms) == 3:
            ci = atoms[2]
            lo, hi = get_lo_hi_from_CI(ci, exclude=exclude)
            if lo > c:
                lo = c
            if hi < c:
                hi = c
            data.append((t, c, lo, hi))
        else:
            data.append((t, c))
    return data


def compute_rmsd(truth, a, limit=150):
    truth = truth[:limit]
    a = a[:limit]
    if len(a) > len(truth):
        a = a[: len(truth)]
    return (sum((i - j) ** 2 for (i, j) in zip(truth, a)) / len(truth)) ** 0.5


def compare(args):
    """
    %prog compare Evaluation.csv

    Compare performances of various variant callers on simulated STR datasets.
    """
    p = OptionParser(compare.__doc__)
    opts, args, iopts = p.set_image_options(args, figsize="10x10")

    if len(args) != 1:
        sys.exit(not p.print_help())

    (datafile,) = args
    pf = datafile.rsplit(".", 1)[0]
    fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(
        ncols=2, nrows=2, figsize=(iopts.w, iopts.h)
    )
    plt.tight_layout(pad=3)

    bbox = {"facecolor": "tomato", "alpha": 0.2, "ec": "w"}
    pad = 2

    # Read benchmark data
    df = pd.read_csv("Evaluation.csv")
    truth = df["Truth"]
    axes = (ax1, ax2, ax3, ax4)
    progs = ("Manta", "Isaac", "GATK", "lobSTR")
    markers = ("bx-", "yo-", "md-", "c+-")

    for ax, prog, marker in zip(axes, progs, markers):
        ax.plot(truth, df[prog], marker)
        ax.plot(truth, truth, "k--")  # to show diagonal
        ax.axhline(infected_thr, color="tomato")
        ax.text(
            max(truth) - pad,
            infected_thr + pad,
            "Risk threshold",
            bbox=bbox,
            ha="right",
        )
        ax.axhline(ref_thr, color="tomato")
        ax.text(
            max(truth) - pad,
            ref_thr - pad,
            "Reference repeat count",
            bbox=bbox,
            ha="right",
            va="top",
        )
        ax.set_title(SIMULATED_HAPLOID)
        ax.set_xlabel(r"Num of CAG repeats inserted ($\mathit{h}$)")
        ax.set_ylabel("Num of CAG repeats called")
        ax.legend([prog, "Truth"], loc="best")

    root = fig.add_axes([0, 0, 1, 1])
    pad = 0.03
    panel_labels(
        root,
        (
            (pad / 2, 1 - pad, "A"),
            (1 / 2.0, 1 - pad, "B"),
            (pad / 2, 1 / 2.0, "C"),
            (1 / 2.0, 1 / 2.0, "D"),
        ),
    )
    normalize_axes(root)

    image_name = pf + "." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def plot_compare(
    ax,
    title,
    tredparse_results,
    lobstr_results,
    pad=8,
    ms=3,
    max_insert=300,
    color="g",
    risk=True,
):
    truth = range(1, max_insert + 1)
    tx, ty, tl, th = zip(*tredparse_results)
    trmsd = compute_rmsd(truth, ty)
    if lobstr_results:
        lx, ly = zip(*lobstr_results)
        lrmsd = compute_rmsd(truth, ly)

    rmsd_tag = "$RMSD_{1:150}$"
    if lobstr_results:
        ax.plot(
            lx, ly, "c+-", ms=ms, label="lobSTR ({}={:.2f})".format(rmsd_tag, lrmsd)
        )
    ax.plot(
        tx,
        ty,
        ".-",
        color=color,
        ms=ms,
        label="TREDPARSE ({}={:.2f})".format(rmsd_tag, trmsd),
    )
    ax.plot(truth, truth, "k--", label="Truth")
    ax.fill_between(
        tx, tl, th, facecolor=color, alpha=0.25, label=r"TREDPARSE 95$\%$ CI"
    )

    ax.set_xlabel(r"Num of CAG repeats inserted ($\mathit{h}$)")
    ax.set_ylabel("Num of CAG repeats called")
    ax.set_title(title)
    ax.legend(loc="best")

    bbox = {"facecolor": "tomato", "alpha": 0.2, "ec": "w"}
    if risk:
        ax.axhline(infected_thr, color="tomato")
        ax.text(
            max(truth) - pad,
            infected_thr + pad,
            "Risk cutoff={}".format(infected_thr) + r"$\times$CAGs",
            bbox=bbox,
            ha="right",
        )
    else:
        readlength, pairdistance = 150 / 3, 500 / 3
        ax.axhline(readlength, color="tomato")
        ax.text(
            max(truth) - pad,
            readlength + pad,
            "Read Length ($L$)",
            bbox=bbox,
            ha="right",
        )
        ax.axhline(pairdistance, color="tomato")
        ax.text(
            max(truth) - pad,
            pairdistance + pad,
            "Paired-end distance($V$)",
            bbox=bbox,
            ha="right",
        )


def compare2(args):
    """
    %prog compare2

    Compare performances of various variant callers on simulated STR datasets.
    """
    p = OptionParser(compare2.__doc__)
    p.add_option(
        "--maxinsert", default=300, type="int", help="Maximum number of repeats"
    )
    add_simulate_options(p)
    opts, args, iopts = p.set_image_options(args, figsize="10x5")

    if len(args) != 0:
        sys.exit(not p.print_help())

    depth = opts.depth
    readlen = opts.readlen
    distance = opts.distance
    max_insert = opts.maxinsert
    fig, (ax1, ax2) = plt.subplots(ncols=2, nrows=1, figsize=(iopts.w, iopts.h))
    plt.tight_layout(pad=2)

    # ax1: lobSTR vs TREDPARSE with haploid model
    lobstr_results = parse_results("lobstr_results_homo.txt")
    tredparse_results = parse_results("tredparse_results_homo.txt")
    title = SIMULATED_HAPLOID + r" ($D=%s\times, L=%dbp, V=%dbp$)" % (
        depth,
        readlen,
        distance,
    )
    plot_compare(ax1, title, tredparse_results, lobstr_results, max_insert=max_insert)

    # ax2: lobSTR vs TREDPARSE with diploid model
    lobstr_results = parse_results("lobstr_results_het.txt", exclude=20)
    tredparse_results = parse_results("tredparse_results_het.txt", exclude=20)
    title = SIMULATED_DIPLOID + r" ($D=%s\times, L=%dbp, V=%dbp$)" % (
        depth,
        readlen,
        distance,
    )
    plot_compare(ax2, title, tredparse_results, lobstr_results, max_insert=max_insert)

    for ax in (ax1, ax2):
        ax.set_xlim(0, max_insert)
        ax.set_ylim(0, max_insert)

    root = fig.add_axes([0, 0, 1, 1])
    pad = 0.03
    panel_labels(root, ((pad / 2, 1 - pad, "A"), (1 / 2.0, 1 - pad, "B")))
    normalize_axes(root)

    image_name = "tredparse." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def power(args):
    """
    %prog power

    Compare performances of various variant callers on simulated STR datasets.
    This compares the power of various evidence types.
    """
    p = OptionParser(power.__doc__)
    p.add_option(
        "--maxinsert", default=300, type="int", help="Maximum number of repeats"
    )
    add_simulate_options(p)
    opts, args, iopts = p.set_image_options(args, figsize="10x10", format="png")

    if len(args) != 0:
        sys.exit(not p.print_help())

    max_insert = opts.maxinsert
    fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(
        ncols=2, nrows=2, figsize=(iopts.w, iopts.h)
    )
    plt.tight_layout(pad=3)

    color = "lightslategray"
    # ax1: Spanning
    tredparse_results = parse_results("tredparse_results_het-spanning.txt")
    title = SIMULATED_DIPLOID + " (Sub-model 1: Spanning reads)"
    plot_compare(
        ax1,
        title,
        tredparse_results,
        None,
        color=color,
        max_insert=max_insert,
        risk=False,
    )

    # ax2: Partial
    tredparse_results = parse_results("tredparse_results_het-partial.txt", exclude=20)
    title = SIMULATED_DIPLOID + " (Sub-model 2: Partial reads)"
    plot_compare(
        ax2,
        title,
        tredparse_results,
        None,
        color=color,
        max_insert=max_insert,
        risk=False,
    )

    # ax3: Repeat
    tredparse_results = parse_results("tredparse_results_het-repeat.txt", exclude=20)
    # HACK (repeat reads won't work under 50)
    tredparse_results = [x for x in tredparse_results if x[0] > 50]
    title = SIMULATED_DIPLOID + " (Sub-model 3: Repeat-only reads)"
    plot_compare(
        ax3,
        title,
        tredparse_results,
        None,
        color=color,
        max_insert=max_insert,
        risk=False,
    )

    # ax4: Pair
    tredparse_results = parse_results("tredparse_results_het-pair.txt", exclude=20)
    title = SIMULATED_DIPLOID + " (Sub-model 4: Paired-end reads)"
    plot_compare(
        ax4,
        title,
        tredparse_results,
        None,
        color=color,
        max_insert=max_insert,
        risk=False,
    )

    for ax in (ax1, ax2, ax3, ax4):
        ax.set_xlim(0, max_insert)
        ax.set_ylim(0, max_insert)

    root = fig.add_axes([0, 0, 1, 1])
    pad = 0.03
    panel_labels(
        root,
        (
            (pad / 2, 1 - pad, "A"),
            (1 / 2.0, 1 - pad, "B"),
            (pad / 2, 1 / 2.0, "C"),
            (1 / 2.0, 1 / 2.0, "D"),
        ),
    )
    normalize_axes(root)

    image_name = "power." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def tredparse(args):
    """
    %prog tredparse

    Compare performances of various variant callers on simulated STR datasets.
    Adds coverage comparisons as panel C and D.
    """
    p = OptionParser(tredparse.__doc__)
    p.add_option(
        "--maxinsert", default=300, type="int", help="Maximum number of repeats"
    )
    add_simulate_options(p)
    opts, args, iopts = p.set_image_options(args, figsize="10x10")

    if len(args) != 0:
        sys.exit(not p.print_help())

    depth = opts.depth
    max_insert = opts.maxinsert
    fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(
        ncols=2, nrows=2, figsize=(iopts.w, iopts.h)
    )
    plt.tight_layout(pad=3)

    # ax1: lobSTR vs TREDPARSE with haploid model
    lobstr_results = parse_results("lobstr_results_homo-20x-150bp-500bp.txt")
    tredparse_results = parse_results("tredparse_results_homo-20x-150bp-500bp.txt")
    title = SIMULATED_HAPLOID + r" (Depth=$%s\times$)" % depth
    plot_compare(ax1, title, tredparse_results, lobstr_results, max_insert=max_insert)

    # ax2: lobSTR vs TREDPARSE with diploid model (depth=20x)
    lobstr_results = parse_results("lobstr_results_het-20x-150bp-500bp.txt", exclude=20)
    tredparse_results = parse_results(
        "tredparse_results_het-20x-150bp-500bp.txt", exclude=20
    )
    title = SIMULATED_DIPLOID + r" (Depth=$%s\times$)" % depth
    plot_compare(ax2, title, tredparse_results, lobstr_results, max_insert=max_insert)

    # ax3: lobSTR vs TREDPARSE with diploid model (depth=5x)
    lobstr_results = parse_results("lobstr_results_het-5x-150bp-500bp.txt", exclude=20)
    tredparse_results = parse_results(
        "tredparse_results_het-5x-150bp-500bp.txt", exclude=20
    )
    title = SIMULATED_DIPLOID + r" (Depth=$%s\times$)" % 5
    plot_compare(ax3, title, tredparse_results, lobstr_results, max_insert=max_insert)

    # ax4: lobSTR vs TREDPARSE with diploid model (depth=80x)
    lobstr_results = parse_results("lobstr_results_het-80x-150bp-500bp.txt", exclude=20)
    tredparse_results = parse_results(
        "tredparse_results_het-80x-150bp-500bp.txt", exclude=20
    )
    title = SIMULATED_DIPLOID + r" (Depth=$%s\times$)" % 80
    plot_compare(ax4, title, tredparse_results, lobstr_results, max_insert=max_insert)

    for ax in (ax1, ax2, ax3, ax4):
        ax.set_xlim(0, max_insert)
        ax.set_ylim(0, max_insert)

    root = fig.add_axes([0, 0, 1, 1])
    pad = 0.03
    panel_labels(
        root,
        (
            (pad / 2, 1 - pad, "A"),
            (1 / 2.0, 1 - pad, "B"),
            (pad / 2, 1 / 2.0, "C"),
            (1 / 2.0, 1 / 2.0, "D"),
        ),
    )
    normalize_axes(root)

    image_name = "tredparse." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


if __name__ == "__main__":
    main()



"""
parses tabular BLAST -m8 (-format 6 in BLAST+) format
"""

import os.path as op
import sys
import logging

from itertools import groupby
from collections import defaultdict

from jcvi.formats.base import LineFile, BaseFile, must_open
from jcvi.formats.bed import Bed
from jcvi.formats.sizes import Sizes
from jcvi.utils.grouper import Grouper
from jcvi.utils.orderedcollections import OrderedDict
from jcvi.utils.range import range_distance
from jcvi.utils.cbook import percentage
from jcvi.assembly.base import calculate_A50
from jcvi.apps.base import OptionParser, ActionDispatcher, sh, popen


try:
    from .cblast import BlastLine
except:
    from .pyblast import BlastLine

    logging.error("Fall back to Python implementation of BlastLine")


class BlastSlow(LineFile):
    """
    Load entire blastfile into memory
    """

    def __init__(self, filename, sorted=False):
        super(BlastSlow, self).__init__(filename)
        fp = must_open(filename)
        for row in fp:
            self.append(BlastLine(row))
        self.sorted = sorted
        if not sorted:
            self.sort(key=lambda x: x.query)

    def iter_hits(self):
        for query, blines in groupby(self, key=lambda x: x.query):
            yield query, blines

    def iter_hits_pair(self):
        key = lambda x: (x.query, x.subject)
        if not self.sorted:
            self.sort(key=key)
        for qs, blines in groupby(self, key=key):
            yield qs, blines

    def to_dict(self):
        # for multiple HSPs pick the one with highest score
        d = OrderedDict()
        for line in self:
            if (line.query, line.subject) not in d:
                d[(line.query, line.subject)] = line
            else:
                cur_score = d[(line.query, line.subject)].score
                if line.score > cur_score:
                    d[(line.query, line.subject)] = line
        return d


class Blast(BaseFile):
    """
    We can have a Blast class that loads entire file into memory, this is
    not very efficient for big files (BlastSlow); when the BLAST file is
    generated by BLAST/BLAT, the file is already sorted
    """

    def __init__(self, filename):
        super(Blast, self).__init__(filename)
        self.fp = must_open(filename)

    def __iter__(self):
        self.fp.seek(0)
        for row in self.fp:
            if row[0] == "#":
                continue
            yield BlastLine(row)

    def iter_hits(self):
        for query, blines in groupby(self.fp, key=lambda x: BlastLine(x).query):
            blines = [BlastLine(x) for x in blines]
            blines.sort(key=lambda x: -x.score)  # descending score
            yield query, blines

    def iter_best_hit(self, N=1, hsps=False, ref="query"):
        if ref == "query":
            ref, hit = "query", "subject"
        elif ref == "subject":
            ref, hit = "subject", "query"
        else:
            sys.exit("`ref` must be either `query` or `subject`.")

        for bref, blines in groupby(self.fp, key=lambda x: getattr(BlastLine(x), ref)):
            blines = [BlastLine(x) for x in blines]
            blines.sort(key=lambda x: -x.score)
            counter = 0
            selected = set()
            for b in blines:
                if hsps:
                    selected.add(getattr(b, hit))
                    counter = len(selected)
                    if counter > N:
                        selected.remove(getattr(b, hit))
                        continue
                else:
                    counter += 1
                    if counter > N:
                        break

                yield bref, b

    @property
    def hits(self):
        """
        returns a dict with query => blastline
        """
        return dict(self.iter_hits())

    @property
    def best_hits(self):
        """
        returns a dict with query => best blasthit
        """
        return dict(self.iter_best_hit())


class BlastLineByConversion(BlastLine):
    """
    make BlastLine object from tab delimited line objects with
    BlastLine-like up to 12 fields formats
    """

    def __init__(self, sline, mode="1" * 12):

        if int(mode, 2) == 4095:
            super(BlastLineByConversion, self).__init__(sline)
        elif 3072 <= int(mode, 2) < 4095:
            args = sline.split("\t")
            atoms = args[:2]
            mode = list(mode)
            if len(args) == 12:
                for i in range(2, 12):
                    if mode[i] == "1":
                        atoms.append(args[i])
                    else:
                        atoms.append("-1")
            if len(args) < 12:
                for i in range(2, 12):
                    if mode[i] == "1":
                        atoms.append(args[i - mode[:i].count("0")])
                    else:
                        atoms.append("-1")
            sline = "\t".join(atoms)
            super(BlastLineByConversion, self).__init__(sline)
        else:
            m = "mode can only contain 0 or 1 \n"
            m += "first two fields (query, subject) cannot be empty"
            sys.exit(m)


class AlignStats:
    """
    Stores the alignment statistics that is used in formats.blast.summary()
    and formats.coords.summary()
    """

    def __init__(
        self, filename, qrycovered, refcovered, qryspan, refspan, identicals, AL50
    ):
        self.filename = filename
        self.qrycovered = qrycovered
        self.refcovered = refcovered
        self.qryspan = qryspan
        self.refspan = refspan
        self.identicals = identicals
        self.AL50 = AL50

    def __str__(self):
        pp = lambda x, d: "{:.2f}".format(x * 100.0 / d)
        return "\t".join(
            str(x)
            for x in (
                self.filename,
                self.identicals,
                self.qrycovered,
                pp(self.identicals, self.qrycovered),
                self.refcovered,
                pp(self.identicals, self.refcovered),
                self.qryspan,
                pp(self.identicals, self.qryspan),
                self.refspan,
                pp(self.identicals, self.refspan),
            )
        )

    def print_stats(self):
        qrycovered = self.qrycovered
        refcovered = self.refcovered
        qryspan = self.qryspan
        refspan = self.refspan
        m0 = "AL50 (>=50% of bases in alignment blocks >= this size): {}".format(
            self.AL50
        )
        m1 = "Query coverage: {}".format(percentage(self.identicals, qrycovered))
        m2 = "Reference coverage: {}".format(percentage(self.identicals, refcovered))
        m3 = "Query span: {}".format(percentage(self.identicals, qryspan))
        m4 = "Reference span: {}".format(percentage(self.identicals, refspan))
        print("\n".join((m0, m1, m2, m3, m4)), file=sys.stderr)


def get_stats(blastfile, strict=False):
    from jcvi.utils.range import range_union, range_span
    from .pyblast import BlastLine

    logging.debug("Report stats on `%s`" % blastfile)
    fp = open(blastfile)
    ref_ivs = []
    qry_ivs = []
    identicals = 0
    ngaps = 0
    alignlens = []

    for row in fp:
        c = BlastLine(row)
        qstart, qstop = c.qstart, c.qstop
        if qstart > qstop:
            qstart, qstop = qstop, qstart
        qry_ivs.append((c.query, qstart, qstop))

        sstart, sstop = c.sstart, c.sstop
        if sstart > sstop:
            sstart, sstop = sstop, sstart
        ref_ivs.append((c.subject, sstart, sstop))

        alen = c.hitlen
        ngaps += c.ngaps
        identicals += c.hitlen - c.nmismatch - c.ngaps
        alignlens.append(alen)

    qrycovered = range_union(qry_ivs)
    refcovered = range_union(ref_ivs)
    if strict:
        # We discount gaps in counting covered bases, since we
        # did not track individually gaps in qry and ref, we assume
        # the gaps are opened evenly in the two sequences
        qrycovered -= ngaps / 2
        refcovered -= ngaps / 2
    qryspan = range_span(qry_ivs)
    refspan = range_span(ref_ivs)
    _, AL50, _ = calculate_A50(alignlens)
    filename = op.basename(blastfile)
    alignstats = AlignStats(
        filename, qrycovered, refcovered, qryspan, refspan, identicals, AL50
    )

    return alignstats


def filter(args):
    """
    %prog filter test.blast

    Produce a new blast file and filter based on:
    - score: >= cutoff
    - pctid: >= cutoff
    - hitlen: >= cutoff
    - evalue: <= cutoff
    - ids: valid ids

    Use --inverse to obtain the complementary records for the criteria above.

    - noself: remove self-self hits
    """
    p = OptionParser(filter.__doc__)
    p.add_option("--score", dest="score", default=0, type="int", help="Score cutoff")
    p.set_align(pctid=95, hitlen=100, evalue=0.01)
    p.add_option(
        "--noself", default=False, action="store_true", help="Remove self-self hits"
    )
    p.add_option("--ids", help="Path to file with ids to retain")
    p.add_option(
        "--inverse",
        default=False,
        action="store_true",
        help="Similar to grep -v, inverse",
    )
    p.set_outfile(outfile=None)

    opts, args = p.parse_args(args)
    if len(args) != 1:
        sys.exit(not p.print_help())

    if opts.ids:
        ids = set()
        for row in must_open(opts.ids):
            if row[0] == "#":
                continue
            row = row.replace(",", "\t")
            ids.update(row.split())
    else:
        ids = None

    (blastfile,) = args
    inverse = opts.inverse
    outfile = opts.outfile
    fp = must_open(blastfile)

    score, pctid, hitlen, evalue, noself = (
        opts.score,
        opts.pctid,
        opts.hitlen,
        opts.evalue,
        opts.noself,
    )
    newblastfile = (
        blastfile + ".P{0}L{1}".format(int(pctid), hitlen)
        if outfile is None
        else outfile
    )
    if inverse:
        newblastfile += ".inverse"
    fw = must_open(newblastfile, "w")
    for row in fp:
        if row[0] == "#":
            continue
        c = BlastLine(row)

        if ids:
            if c.query in ids and c.subject in ids:
                noids = False
            else:
                noids = True
        else:
            noids = None

        remove = (
            c.score < score
            or c.pctid < pctid
            or c.hitlen < hitlen
            or c.evalue > evalue
            or noids
        )

        if inverse:
            remove = not remove

        remove = remove or (noself and c.query == c.subject)

        if not remove:
            print(row.rstrip(), file=fw)

    fw.close()

    return newblastfile


def main():

    actions = (
        ("summary", "provide summary on id% and cov%"),
        ("completeness", "print completeness statistics for each query"),
        ("annotation", "create tabular file with the annotations"),
        ("top10", "count the most frequent 10 hits"),
        ("filter", "filter BLAST file (based on score, id%, alignlen)"),
        ("covfilter", "filter BLAST file (based on id% and cov%)"),
        ("cscore", "calculate C-score for BLAST pairs"),
        ("best", "get best BLAST hit per query"),
        ("pairs", "print paired-end reads of BLAST tabular file"),
        ("bed", "get bed file from BLAST tabular file"),
        ("condense", "group HSPs together for same query-subject pair"),
        ("chain", "chain adjacent HSPs together"),
        ("swap", "swap query and subjects in BLAST tabular file"),
        ("sort", "sort lines so that query grouped together and scores desc"),
        ("subset", "extract hits from some query and subject chrs"),
        ("mismatches", "print out histogram of mismatches of HSPs"),
        ("annotate", "annotate overlap types in BLAST tabular file"),
        ("score", "add up the scores for each query seq"),
        ("rbbh", "find reciprocal-best blast hits"),
        ("gaps", "find distribution of gap sizes between adjacent HSPs"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def collect_gaps(blast, use_subject=False):
    """
    Collect the gaps between adjacent HSPs in the BLAST file.
    """
    key = lambda x: x.sstart if use_subject else x.qstart
    blast.sort(key=key)

    for a, b in zip(blast, blast[1:]):
        if use_subject:
            if a.sstop < b.sstart:
                yield b.sstart - a.sstop
        else:
            if a.qstop < b.qstart:
                yield b.qstart - a.qstop


def gaps(args):
    """
    %prog gaps A_vs_B.blast

    Find distribution of gap sizes betwen adjacent HSPs.
    """
    p = OptionParser(gaps.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = args
    blast = BlastSlow(blastfile)
    logging.debug("A total of {} records imported".format(len(blast)))

    query_gaps = list(collect_gaps(blast))
    subject_gaps = list(collect_gaps(blast, use_subject=True))
    logging.debug(
        "Query gaps: {}  Subject gaps: {}".format(len(query_gaps), len(subject_gaps))
    )

    from jcvi.graphics.base import savefig
    import seaborn as sns

    sns.distplot(query_gaps)
    savefig("query_gaps.pdf")


def rbbh(args):
    """
    %prog rbbh A_vs_B.blast B_vs_A.blast

    Identify the reciprocal best blast hit for each query sequence in set A
    when compared to set B.

    This program assumes that the BLAST results have already been filtered
    based on a combination of %id, %cov, e-value cutoffs. BLAST output should
    be in tabular `-m 8` format.
    """
    p = OptionParser(rbbh.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    (
        abfile,
        bafile,
    ) = args
    ab = Blast(abfile)
    ba = Blast(bafile)

    ab_hits = ab.best_hits
    ba_hits = ba.best_hits

    for aquery in ab_hits:
        ahit = ab_hits[aquery].subject
        ba_bline = ba_hits.get(ahit)
        if ba_bline:
            bhit = ba_bline.subject
            if bhit == aquery:
                print("\t".join(str(x) for x in (aquery, ahit)))


def score(args):
    """
    %prog score blastfile query.fasta A.ids

    Add up the scores for each query seq. Go through the lines and for each
    query sequence, add up the scores when subject is in each pile by A.ids.
    """
    from jcvi.formats.base import SetFile
    from jcvi.formats.fasta import Fasta

    p = OptionParser(score.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 3:
        sys.exit(not p.print_help())

    blastfile, fastafile, idsfile = args
    ids = SetFile(idsfile)

    blast = Blast(blastfile)
    scores = defaultdict(int)
    for b in blast:
        query = b.query
        subject = b.subject
        if subject not in ids:
            continue
        scores[query] += b.score

    logging.debug("A total of {0} ids loaded.".format(len(ids)))

    f = Fasta(fastafile)
    for s in f.iterkeys_ordered():
        sc = scores.get(s, 0)
        print("\t".join((s, str(sc))))


def annotation(args):
    """
    %prog annotation blastfile > annotations

    Create simple two column files from the first two coluns in blastfile. Use
    --queryids and --subjectids to switch IDs or descriptions.
    """
    from jcvi.formats.base import DictFile

    p = OptionParser(annotation.__doc__)
    p.add_option("--queryids", help="Query IDS file to switch")
    p.add_option("--subjectids", help="Subject IDS file to switch")
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = args

    d = "\t"
    qids = DictFile(opts.queryids, delimiter=d) if opts.queryids else None
    sids = DictFile(opts.subjectids, delimiter=d) if opts.subjectids else None
    blast = Blast(blastfile)
    for b in blast:
        query, subject = b.query, b.subject
        if qids:
            query = qids[query]
        if sids:
            subject = sids[subject]
        print("\t".join((query, subject)))


def completeness(args):
    """
    %prog completeness blastfile ref.fasta > outfile

    Print statistics for each gene, the coverage of the alignment onto the best hit,
    as an indicator for completeness of the gene model. For example, one might
    BLAST sugarcane ESTs against sorghum annotations as reference, to find
    full-length transcripts.
    """
    from jcvi.utils.range import range_minmax
    from jcvi.utils.cbook import SummaryStats

    p = OptionParser(completeness.__doc__)
    p.add_option("--ids", help="Save ids that are over 50% complete")
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    blastfile, fastafile = args
    idsfile = opts.ids
    f = Sizes(fastafile).mapping

    b = BlastSlow(blastfile)
    valid = []
    data = []
    cutoff = 50
    for query, blines in groupby(b, key=lambda x: x.query):
        blines = list(blines)
        ranges = [(x.sstart, x.sstop) for x in blines]
        b = blines[0]
        query, subject = b.query, b.subject

        rmin, rmax = range_minmax(ranges)
        subject_len = f[subject]

        nterminal_dist = rmin - 1
        cterminal_dist = subject_len - rmax
        covered = (rmax - rmin + 1) * 100 / subject_len
        if covered > cutoff:
            valid.append(query)

        data.append((nterminal_dist, cterminal_dist, covered))
        print(
            "\t".join(
                str(x)
                for x in (query, subject, nterminal_dist, cterminal_dist, covered)
            )
        )

    nd, cd, cv = zip(*data)
    m = "Total: {0}, Coverage > {1}%: {2}\n".format(len(data), cutoff, len(valid))
    m += "N-terminal: {0}\n".format(SummaryStats(nd))
    m += "C-terminal: {0}\n".format(SummaryStats(cd))
    m += "Coverage: {0}".format(SummaryStats(cv))
    print(m, file=sys.stderr)

    if idsfile:
        fw = open(idsfile, "w")
        print("\n".join(valid), file=fw)
        logging.debug(
            "A total of {0} ids (cov > {1} %) written to `{2}`.".format(
                len(valid), cutoff, idsfile
            )
        )
        fw.close()


def annotate(args):
    """
    %prog annotate blastfile query.fasta subject.fasta

    Annotate overlap types (dovetail, contained, etc) in BLAST tabular file.
    """
    from jcvi.assembly.goldenpath import Cutoff, Overlap, Overlap_types

    p = OptionParser(annotate.__doc__)
    p.set_align(pctid=94, hitlen=500)
    p.add_option("--hang", default=500, type="int", help="Maximum overhang length")
    opts, args = p.parse_args(args)

    if len(args) != 3:
        sys.exit(not p.print_help())

    blastfile, afasta, bfasta = args
    fp = must_open(blastfile)
    asizes = Sizes(afasta).mapping
    bsizes = Sizes(bfasta).mapping
    cutoff = Cutoff(opts.pctid, opts.hitlen, opts.hang)
    logging.debug(str(cutoff))
    for row in fp:
        b = BlastLine(row)
        asize = asizes[b.query]
        bsize = bsizes[b.subject]
        if b.query == b.subject:
            continue
        ov = Overlap(b, asize, bsize, cutoff)
        if ov.otype:
            ov.print_graphic()
            print("{0}\t{1}".format(b, Overlap_types[ov.otype]))


def top10(args):
    """
    %prog top10 blastfile.best

    Count the most frequent 10 hits. Usually the BLASTFILE needs to be screened
    the get the best match. You can also provide an .ids file to query the ids.
    For example the ids file can contain the seqid to species mapping.

    The ids file is two-column, and can sometimes be generated by
    `jcvi.formats.fasta ids --description`.
    """
    from jcvi.formats.base import DictFile

    p = OptionParser(top10.__doc__)
    p.add_option(
        "--top",
        default=10,
        type="int",
        help="Top N taxa to extract",
    )
    p.add_option(
        "--ids",
        default=None,
        help="Two column ids file to query seqid",
    )
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = args
    mapping = DictFile(opts.ids, delimiter="\t") if opts.ids else {}

    cmd = "cut -f2 {0}".format(blastfile)
    cmd += " | sort | uniq -c | sort -k1,1nr | head -n {0}".format(opts.top)
    fp = popen(cmd)
    for row in fp:
        count, seqid = row.split()
        nseqid = mapping.get(seqid, seqid)
        print("\t".join((count, nseqid)))


def sort(args):
    """
    %prog sort <blastfile|coordsfile>

    Sort lines so that same query grouped together with scores descending. The
    sort is 'in-place'.
    """
    p = OptionParser(sort.__doc__)
    p.add_option(
        "--query",
        default=False,
        action="store_true",
        help="Sort by query position",
    )
    p.add_option(
        "--ref",
        default=False,
        action="store_true",
        help="Sort by reference position",
    )
    p.add_option(
        "--refscore",
        default=False,
        action="store_true",
        help="Sort by reference name, then score descending",
    )
    p.add_option(
        "--coords",
        default=False,
        action="store_true",
        help="File is .coords generated by NUCMER",
    )
    p.set_tmpdir()

    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = args

    if opts.coords:
        if opts.query:
            key = "-k13,13 -k3,3n"
        elif opts.ref:
            key = "-k12,12 -k1,1n"

    else:
        if opts.query:
            key = "-k1,1 -k7,7n"
        elif opts.ref:
            key = "-k2,2 -k9,9n"
        elif opts.refscore:
            key = "-k2,2 -k12,12gr"
        else:
            key = "-k1,1 -k12,12gr"

    cmd = "sort"
    if opts.tmpdir:
        cmd += " -T {0}".format(opts.tmpdir)
    cmd += " {0} {1} -o {1}".format(key, blastfile)
    sh(cmd)


def cscore(args):
    """
    %prog cscore blastfile > cscoreOut

    See supplementary info for sea anemone genome paper, C-score formula:

        cscore(A,B) = score(A,B) /
             max(best score for A, best score for B)

    A C-score of one is the same as reciprocal best hit (RBH).

    Output file will be 3-column (query, subject, cscore). Use --cutoff to
    select a different cutoff.
    """
    from jcvi.utils.cbook import gene_name

    p = OptionParser(cscore.__doc__)
    p.add_option(
        "--cutoff",
        default=0.9999,
        type="float",
        help="Minimum C-score to report",
    )
    p.add_option(
        "--pct",
        default=False,
        action="store_true",
        help="Also include pct as last column",
    )
    p.add_option(
        "--writeblast",
        default=False,
        action="store_true",
        help="Also write filtered blast file",
    )
    p.set_stripnames()
    p.set_outfile()

    opts, args = p.parse_args(args)
    ostrip = opts.strip_names
    writeblast = opts.writeblast
    outfile = opts.outfile

    if len(args) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = args

    blast = Blast(blastfile)
    logging.debug("Register best scores ..")
    best_score = defaultdict(float)
    for b in blast:
        query, subject = b.query, b.subject
        if ostrip:
            query, subject = gene_name(query), gene_name(subject)

        score = b.score
        if score > best_score[query]:
            best_score[query] = score
        if score > best_score[subject]:
            best_score[subject] = score

    blast = Blast(blastfile)
    pairs = {}
    cutoff = opts.cutoff
    for b in blast:
        query, subject = b.query, b.subject
        if ostrip:
            query, subject = gene_name(query), gene_name(subject)

        score = b.score
        pctid = b.pctid
        s = score / max(best_score[query], best_score[subject])
        if s > cutoff:
            pair = (query, subject)
            if pair not in pairs or s > pairs[pair][0]:
                pairs[pair] = (s, pctid, b)

    fw = must_open(outfile, "w")
    if writeblast:
        fwb = must_open(outfile + ".filtered.blast", "w")
    pct = opts.pct
    for (query, subject), (s, pctid, b) in sorted(pairs.items()):
        args = [query, subject, "{0:.2f}".format(s)]
        if pct:
            args.append("{0:.1f}".format(pctid))
        print("\t".join(args), file=fw)
        if writeblast:
            print(b, file=fwb)
    fw.close()
    if writeblast:
        fwb.close()


def get_distance(a, b, xaxis=True):
    """
    Returns the distance between two blast HSPs.
    """
    if xaxis:
        arange = ("0", a.qstart, a.qstop, a.orientation)  # 0 is the dummy chromosome
        brange = ("0", b.qstart, b.qstop, b.orientation)
    else:
        arange = ("0", a.sstart, a.sstop, a.orientation)
        brange = ("0", b.sstart, b.sstop, b.orientation)

    dist, oo = range_distance(arange, brange, distmode="ee")
    dist = abs(dist)

    return dist


def combine_HSPs(a):
    """
    Combine HSPs into a single BlastLine.
    """
    m = a[0]
    if len(a) == 1:
        return m

    for b in a[1:]:
        assert m.query == b.query
        assert m.subject == b.subject
        m.hitlen += b.hitlen
        m.nmismatch += b.nmismatch
        m.ngaps += b.ngaps
        m.qstart = min(m.qstart, b.qstart)
        m.qstop = max(m.qstop, b.qstop)
        m.sstart = min(m.sstart, b.sstart)
        m.sstop = max(m.sstop, b.sstop)
        if m.has_score:
            m.score += b.score

    m.pctid = 100 - (m.nmismatch + m.ngaps) * 100.0 / m.hitlen
    return m


def chain_HSPs(blast, xdist=100, ydist=100):
    """
    Take a list of BlastLines (or a BlastSlow instance), and returns a list of
    BlastLines.
    """
    key = lambda x: (x.query, x.subject)
    blast.sort(key=key)

    clusters = Grouper()
    for qs, points in groupby(blast, key=key):
        points = sorted(
            list(points), key=lambda x: (x.qstart, x.qstop, x.sstart, x.sstop)
        )

        n = len(points)
        for i in range(n):
            a = points[i]
            clusters.join(a)
            for j in range(i + 1, n):
                b = points[j]

                # x-axis distance
                del_x = get_distance(a, b)
                if del_x > xdist:
                    break
                # y-axis distance
                del_y = get_distance(a, b, xaxis=False)
                if del_y > ydist:
                    continue
                # otherwise join
                clusters.join(a, b)

    chained_hsps = [combine_HSPs(x) for x in clusters]
    key = lambda x: (x.query, -x.score if x.has_score else 0)
    chained_hsps = sorted(chained_hsps, key=key)

    return chained_hsps


def chain(args):
    """
    %prog chain blastfile

    Chain adjacent HSPs together to form larger HSP.
    """
    p = OptionParser(chain.__doc__)
    p.add_option(
        "--dist",
        dest="dist",
        default=100,
        type="int",
        help="extent of flanking regions to search",
    )

    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = args
    dist = opts.dist
    assert dist > 0

    blast = BlastSlow(blastfile)
    logging.debug("A total of {} records imported".format(len(blast)))
    chained_hsps = chain_HSPs(blast, xdist=dist, ydist=dist)
    logging.debug("A total of {} records after chaining".format(len(chained_hsps)))

    for b in chained_hsps:
        print(b)


def condense(args):
    """
    %prog condense blastfile > blastfile.condensed

    Condense HSPs that belong to the same query-subject pair into one.
    """
    p = OptionParser(condense.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = args
    blast = BlastSlow(blastfile)
    key = lambda x: x.query
    blast.sort(key=key)

    clusters = []
    for q, lines in groupby(blast, key=key):
        lines = list(lines)
        condenser = defaultdict(list)

        for b in lines:
            condenser[(b.subject, b.orientation)].append(b)

        for bs in condenser.values():
            clusters.append(bs)

    chained_hsps = [combine_HSPs(x) for x in clusters]
    chained_hsps = sorted(chained_hsps, key=lambda x: (x.query, -x.score))
    for b in chained_hsps:
        print(b)


def mismatches(args):
    """
    %prog mismatches blastfile

    Print out histogram of mismatches of HSPs, usually for evaluating SNP level.
    """
    from jcvi.utils.cbook import percentage
    from jcvi.graphics.histogram import stem_leaf_plot

    p = OptionParser(mismatches.__doc__)

    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = args

    data = []
    b = Blast(blastfile)
    for query, bline in b.iter_best_hit():
        mm = bline.nmismatch + bline.ngaps
        data.append(mm)

    nonzeros = [x for x in data if x != 0]
    title = "Polymorphic sites: {0}".format(percentage(len(nonzeros), len(data)))
    stem_leaf_plot(data, 0, 20, 20, title=title)


def covfilter(args):
    """
    %prog covfilter blastfile fastafile

    Fastafile is used to get the sizes of the queries. Two filters can be
    applied, the id% and cov%.
    """
    from jcvi.algorithms.supermap import supermap
    from jcvi.utils.range import range_union

    allowed_iterby = ("query", "query_sbjct")

    p = OptionParser(covfilter.__doc__)
    p.set_align(pctid=95, pctcov=50)
    p.add_option(
        "--scov",
        default=False,
        action="store_true",
        help="Subject coverage instead of query",
    )
    p.add_option(
        "--supermap", action="store_true", help="Use supermap instead of union"
    )
    p.add_option(
        "--ids",
        dest="ids",
        default=None,
        help="Print out the ids that satisfy",
    )
    p.add_option(
        "--list",
        dest="list",
        default=False,
        action="store_true",
        help="List the id% and cov% per gene",
    )
    p.add_option(
        "--iterby",
        dest="iterby",
        default="query",
        choices=allowed_iterby,
        help="Choose how to iterate through BLAST",
    )
    p.set_outfile(outfile=None)

    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    blastfile, fastafile = args
    pctid = opts.pctid
    pctcov = opts.pctcov
    union = not opts.supermap
    scov = opts.scov
    sz = Sizes(fastafile)
    sizes = sz.mapping
    iterby = opts.iterby
    qspair = iterby == "query_sbjct"

    if not union:
        querysupermap = blastfile + ".query.supermap"
        if not op.exists(querysupermap):
            supermap(blastfile, filter="query")

        blastfile = querysupermap

    assert op.exists(blastfile)

    covered = 0
    mismatches = 0
    gaps = 0
    alignlen = 0
    queries = set()
    valid = set()
    blast = BlastSlow(blastfile)
    iterator = blast.iter_hits_pair if qspair else blast.iter_hits

    covidstore = {}
    for query, blines in iterator():
        blines = list(blines)
        queries.add(query)

        # per gene report
        this_covered = 0
        this_alignlen = 0
        this_mismatches = 0
        this_gaps = 0
        this_identity = 0

        ranges = []
        for b in blines:
            if scov:
                s, start, stop = b.subject, b.sstart, b.sstop
            else:
                s, start, stop = b.query, b.qstart, b.qstop
            cov_id = s

            if b.pctid < pctid:
                continue

            if start > stop:
                start, stop = stop, start
            this_covered += stop - start + 1
            this_alignlen += b.hitlen
            this_mismatches += b.nmismatch
            this_gaps += b.ngaps
            ranges.append(("1", start, stop))

        if ranges:
            this_identity = (
                100.0 - (this_mismatches + this_gaps) * 100.0 / this_alignlen
            )

        if union:
            this_covered = range_union(ranges)

        this_coverage = this_covered * 100.0 / sizes[cov_id]
        covidstore[query] = (this_identity, this_coverage)
        if this_identity >= pctid and this_coverage >= pctcov:
            valid.add(query)

        covered += this_covered
        mismatches += this_mismatches
        gaps += this_gaps
        alignlen += this_alignlen

    if opts.list:
        if qspair:
            allpairs = defaultdict(list)
            for (q, s) in covidstore:
                allpairs[q].append((q, s))
                allpairs[s].append((q, s))

            for id, size in sz.iter_sizes():
                if id not in allpairs:
                    print("\t".join((id, "na", "0", "0")))
                else:
                    for qs in allpairs[id]:
                        this_identity, this_coverage = covidstore[qs]
                        print(
                            "{0}\t{1:.1f}\t{2:.1f}".format(
                                "\t".join(qs), this_identity, this_coverage
                            )
                        )
        else:
            for query, size in sz.iter_sizes():
                this_identity, this_coverage = covidstore.get(query, (0, 0))
                print(
                    "{0}\t{1:.1f}\t{2:.1f}".format(query, this_identity, this_coverage)
                )

    mapped_count = len(queries)
    valid_count = len(valid)
    cutoff_message = "(id={0.pctid}% cov={0.pctcov}%)".format(opts)

    m = "Identity: {0} mismatches, {1} gaps, {2} alignlen\n".format(
        mismatches, gaps, alignlen
    )
    total = len(sizes.keys())
    m += "Total mapped: {0} ({1:.1f}% of {2})\n".format(
        mapped_count, mapped_count * 100.0 / total, total
    )
    m += "Total valid {0}: {1} ({2:.1f}% of {3})\n".format(
        cutoff_message, valid_count, valid_count * 100.0 / total, total
    )
    m += "Average id = {0:.2f}%\n".format(100 - (mismatches + gaps) * 100.0 / alignlen)

    queries_combined = sz.totalsize
    m += "Coverage: {0} covered, {1} total\n".format(covered, queries_combined)
    m += "Average coverage = {0:.2f}%".format(covered * 100.0 / queries_combined)

    logfile = blastfile + ".covfilter.log"
    fw = open(logfile, "w")
    for f in (sys.stderr, fw):
        print(m, file=f)
    fw.close()

    if opts.ids:
        filename = opts.ids
        fw = must_open(filename, "w")
        for id in valid:
            print(id, file=fw)
        logging.debug(
            "Queries beyond cutoffs {0} written to `{1}`.".format(
                cutoff_message, filename
            )
        )

    outfile = opts.outfile
    if not outfile:
        return

    fw = must_open(outfile, "w")
    blast = Blast(blastfile)
    for b in blast:
        query = (b.query, b.subject) if qspair else b.query
        if query in valid:
            print(b, file=fw)


def swap(args):
    """
    %prog swap blastfile

    Print out a new blast file with query and subject swapped.
    """
    p = OptionParser(swap.__doc__)

    opts, args = p.parse_args(args)

    if len(args) < 1:
        sys.exit(not p.print_help())

    (blastfile,) = args
    swappedblastfile = blastfile + ".swapped"
    fp = must_open(blastfile)
    fw = must_open(swappedblastfile, "w")
    for row in fp:
        b = BlastLine(row)
        print(b.swapped, file=fw)

    fw.close()
    sort([swappedblastfile])


def bed(args):
    """
    %prog bed blastfile

    Print out bed file based on coordinates in BLAST report. By default, write
    out subject positions. Use --swap to write query positions.
    """
    from jcvi.formats.bed import sort as bed_sort

    p = OptionParser(bed.__doc__)
    p.add_option(
        "--swap",
        default=False,
        action="store_true",
        help="Write query positions",
    )
    p.add_option(
        "--both",
        default=False,
        action="store_true",
        help="Generate one line for each of query and subject",
    )

    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(p.print_help())

    (blastfile,) = args
    positive = (not opts.swap) or opts.both
    negative = opts.swap or opts.both

    fp = must_open(blastfile)
    bedfile = (
        "{0}.bed".format(blastfile.rsplit(".", 1)[0])
        if blastfile.endswith(".blast")
        else "{0}.bed".format(blastfile)
    )
    fw = open(bedfile, "w")
    for row in fp:
        b = BlastLine(row)
        if positive:
            print(b.bedline, file=fw)
        if negative:
            print(b.swapped.bedline, file=fw)

    logging.debug("File written to `%s`.", bedfile)
    fw.close()
    bed_sort([bedfile, "-i"])

    return bedfile


def pairs(args):
    """
    See __doc__ for OptionParser.set_pairs().
    """
    import jcvi.formats.bed

    p = OptionParser(pairs.__doc__)
    p.set_pairs()
    opts, targs = p.parse_args(args)

    if len(targs) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = targs
    bedfile = bed([blastfile])
    args[args.index(blastfile)] = bedfile

    return jcvi.formats.bed.pairs(args)


def best(args):
    """
    %prog best blastfile

    print the best hit for each query in the blastfile
    """
    p = OptionParser(best.__doc__)

    p.add_option("-n", default=1, type="int", help="get best N hits")
    p.add_option(
        "--nosort",
        default=False,
        action="store_true",
        help="assume BLAST is already sorted",
    )
    p.add_option(
        "--hsps",
        default=False,
        action="store_true",
        help="get all HSPs for the best pair",
    )
    p.add_option(
        "--subject",
        default=False,
        action="store_true",
        help="get best hit(s) for subject genome instead",
    )
    p.set_tmpdir()
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = args
    n = opts.n
    hsps = opts.hsps
    tmpdir = opts.tmpdir
    ref = "query" if not opts.subject else "subject"

    if not opts.nosort:
        sargs = [blastfile]
        if tmpdir:
            sargs += ["-T {0}".format(tmpdir)]
        if ref != "query":
            sargs += ["--refscore"]
        sort(sargs)
    else:
        logging.debug("Assuming sorted BLAST")

    if not opts.subject:
        bestblastfile = blastfile + ".best"
    else:
        bestblastfile = blastfile + ".subject.best"
    fw = open(bestblastfile, "w")

    b = Blast(blastfile)
    for q, bline in b.iter_best_hit(N=n, hsps=hsps, ref=ref):
        print(bline, file=fw)

    return bestblastfile


def summary(args):
    """
    %prog summary blastfile

    Provide summary on id% and cov%, for both query and reference. Often used in
    comparing genomes (based on NUCMER results).
    """
    p = OptionParser(summary.__doc__)
    p.add_option(
        "--strict",
        default=False,
        action="store_true",
        help="Strict 'gapless' mode. Exclude gaps from covered base.",
    )
    p.add_option(
        "--tabular",
        default=False,
        action="store_true",
        help="Print succint tabular output",
    )

    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = args

    alignstats = get_stats(blastfile, strict=opts.strict)
    if opts.tabular:
        print(str(alignstats))
    else:
        alignstats.print_stats()


def subset(args):
    """
    %prog subset blastfile qbedfile sbedfile

    Extract blast hits between given query and subject chrs.

    If --qchrs or --schrs is not given, then all chrs from q/s genome will
    be included. However one of --qchrs and --schrs must be specified.
    Otherwise the script will do nothing.
    """
    p = OptionParser(subset.__doc__)
    p.add_option(
        "--qchrs",
        default=None,
        help="query chrs to extract, comma sep",
    )
    p.add_option(
        "--schrs",
        default=None,
        help="subject chrs to extract, comma sep",
    )
    p.add_option(
        "--convert",
        default=False,
        action="store_true",
        help="convert accns to chr_rank",
    )
    opts, args = p.parse_args(args)

    if len(args) != 3:
        sys.exit(not p.print_help())

    blastfile, qbedfile, sbedfile = args
    qchrs = opts.qchrs
    schrs = opts.schrs
    assert qchrs or schrs, p.print_help()
    convert = opts.convert

    outfile = blastfile + "."
    if qchrs:
        outfile += qchrs + "."
        qchrs = set(qchrs.split(","))
    else:
        qchrs = set(Bed(qbedfile).seqids)
    if schrs:
        schrs = set(schrs.split(","))
        if qbedfile != sbedfile or qchrs != schrs:
            outfile += ",".join(schrs) + "."
    else:
        schrs = set(Bed(sbedfile).seqids)
    outfile += "blast"

    qo = Bed(qbedfile).order
    so = Bed(sbedfile).order

    fw = must_open(outfile, "w")
    for b in Blast(blastfile):
        q, s = b.query, b.subject
        if qo[q][1].seqid in qchrs and so[s][1].seqid in schrs:
            if convert:
                b.query = qo[q][1].seqid + "_" + "{0:05d}".format(qo[q][0])
                b.subject = so[s][1].seqid + "_" + "{0:05d}".format(so[s][0])
            print(b, file=fw)
    fw.close()
    logging.debug("Subset blastfile written to `{0}`".format(outfile))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Python implementation of BlastLine, an alternative Cython implementation is
available in .cblast.BlastLine, which may be up to 2x faster
"""


class BlastLine(object):
    __slots__ = (
        "query",
        "subject",
        "pctid",
        "hitlen",
        "nmismatch",
        "ngaps",
        "qstart",
        "qstop",
        "sstart",
        "sstop",
        "evalue",
        "score",
        "qseqid",
        "sseqid",
        "qi",
        "si",
        "orientation",
    )

    def __init__(self, sline):
        args = sline.split("\t")
        self.query = args[0]
        self.subject = args[1]
        self.pctid = float(args[2])
        self.hitlen = int(args[3])
        self.nmismatch = int(args[4])
        self.ngaps = int(args[5])
        self.qstart = int(args[6])
        self.qstop = int(args[7])
        self.sstart = int(args[8])
        self.sstop = int(args[9])
        if len(args) > 10:
            self.evalue = float(args[10])
            self.score = float(args[11])

        self.orientation = "+"
        if self.qstart > self.qstop:
            self.qstart, self.qstop = self.qstop, self.qstart
            self.orientation = "-"
        if self.sstart > self.sstop:
            self.sstart, self.sstop = self.sstop, self.sstart
            self.orientation = "-"

    @property
    def has_score(self):
        return hasattr(self, "score")

    def __repr__(self):
        return "BlastLine('%s' to '%s', eval=%.3f, score=%.1f)" % (
            self.query,
            self.subject,
            self.evalue,
            self.score,
        )

    def __str__(self):
        if self.has_score:
            args = [getattr(self, attr) for attr in BlastLine.__slots__[:12]]
        else:
            args = [getattr(self, attr) for attr in BlastLine.__slots__[:10]]
        if self.orientation == "-":
            args[8], args[9] = args[9], args[8]
        return "\t".join(str(x) for x in args)

    @property
    def swapped(self):
        """
        Swap query and subject.
        """
        args = [getattr(self, attr) for attr in BlastLine.__slots__[:12]]
        args[0:2] = [self.subject, self.query]
        args[6:10] = [self.sstart, self.sstop, self.qstart, self.qstop]
        if self.orientation == "-":
            args[8], args[9] = args[9], args[8]
        b = "\t".join(str(x) for x in args)
        return BlastLine(b)

    @property
    def bedline(self):
        return "\t".join(
            str(x)
            for x in (
                self.subject,
                self.sstart - 1,
                self.sstop,
                self.query,
                self.score,
                self.orientation,
            )
        )



#!/usr/bin/env python
# -*- coding: UTF-8 -*-


import sys
import logging

from jcvi.formats.base import write_file
from jcvi.apps.base import OptionParser


default_template = """
\"\"\"

\"\"\"

import sys

{}
from jcvi.apps.base import OptionParser, ActionDispatcher


def main():

    actions = (
        ('app', ''),
            )
    p = ActionDispatcher(actions)
    p.dispatch(globals())

{}

if __name__ == '__main__':
    main()
"""

default_imports = ""

graphic_imports = """
from jcvi.graphics.base import plt, savefig, normalize_axes"""

default_app = """
def app(args):
    \"\"\"
    %prog app datafile

    \"\"\"
    p = OptionParser(app.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())
"""

graphic_app = """
def app(args):
    \"\"\"
    %prog app datafile

    \"\"\"
    p = OptionParser(app.__doc__)
    opts, args, iopts = p.set_image_options(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    datafile, = args
    pf = datafile.rsplit(".", 1)[0]
    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])

    normalize_axes(root)

    image_name = pf + "." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)
"""


def main():
    """
    %prog scriptname.py

    create a minimal boilerplate for a new script
    """
    p = OptionParser(main.__doc__)
    p.add_option(
        "-g",
        "--graphic",
        default=False,
        action="store_true",
        help="Create boilerplate for a graphic script",
    )

    opts, args = p.parse_args()
    if len(args) != 1:
        sys.exit(not p.print_help())

    (script,) = args
    imports = graphic_imports if opts.graphic else default_imports
    app = graphic_app if opts.graphic else default_app
    template = default_template.format(imports, app)
    write_file(script, template)

    message = "template writes to `{0}`".format(script)
    if opts.graphic:
        message = "graphic " + message
    message = message[0].upper() + message[1:]
    logging.debug(message)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import logging
import sys

from collections import defaultdict
from itertools import groupby

from ete3 import Tree

from jcvi.apps.base import OptionParser, glob
from jcvi.formats.base import LineFile
from jcvi.formats.sizes import Sizes
from jcvi.graphics.base import (
    FancyBboxPatch,
    Rectangle,
    linear_shade,
    markup,
    normalize_axes,
    plt,
    savefig,
    set3_n,
)
from jcvi.graphics.glyph import ExonGlyph, TextCircle, get_setups


class LeafInfoLine:
    def __init__(self, row, delimiter=","):
        args = [x.strip() for x in row.split(delimiter)]
        self.name = args[0]
        self.color = args[1]
        self.new_name = None
        if len(args) > 2:
            self.new_name = args[2]


class LeafInfoFile(LineFile):
    def __init__(self, filename, delimiter=","):
        super(LeafInfoFile, self).__init__(filename)
        self.cache = {}
        with open(filename) as fp:
            for row in fp:
                if row[0] == "#":
                    continue
                line = LeafInfoLine(row, delimiter=delimiter)
                self.cache[line.name] = line


class WGDInfoLine:
    def __init__(self, row, delimiter=",", defaultcolor="#7fc97f"):
        args = [x.strip() for x in row.split(delimiter)]
        self.node_name = args[0]
        self.divergence = float(args[1]) / 100
        self.name = args[2]
        self.color = args[3] or defaultcolor
        self.style = args[4]


class WGDInfoFile(LineFile):
    def __init__(self, filename, delimiter=","):
        super(WGDInfoFile, self).__init__(filename)
        self.cache = defaultdict(list)
        with open(filename) as fp:
            for row in fp:
                if row[0] == "#":
                    continue
                line = WGDInfoLine(row, delimiter=delimiter)
                self.cache[line.node_name].append(line)


def truncate_name(name, rule=None):
    """
    shorten taxa names for tree display

    Options of rule. This only affects tree display.
    - headn (eg. head3 truncates first 3 chars)
    - oheadn (eg. ohead3 retains only the first 3 chars)
    - tailn (eg. tail3 truncates last 3 chars)
    - otailn (eg. otail3 retains only the last 3 chars)
    n = 1 ~ 99
    """
    import re

    if rule is None:
        return name

    k = re.search("(?<=^head)[0-9]{1,2}$", rule)
    if k:
        k = k.group(0)
        tname = name[int(k) :]
    else:
        k = re.search("(?<=^ohead)[0-9]{1,2}$", rule)
        if k:
            k = k.group(0)
            tname = name[: int(k)]
        else:
            k = re.search("(?<=^tail)[0-9]{1,2}$", rule)
            if k:
                k = k.group(0)
                tname = name[: -int(k)]
            else:
                k = re.search("(?<=^otail)[0-9]{1,2}$", rule)
                if k:
                    k = k.group(0)
                    tname = name[-int(k) :]
                else:
                    print(truncate_name.__doc__, file=sys.stderr)
                    raise ValueError("Wrong rule for truncation!")
    return tname


def draw_wgd_xy(ax, xx, yy, wgdline):
    """Draw WGD at (xx, yy) position

    Args:
        ax (axis): Matplotlib axes
        xx (float): x position
        yy (float): y position
        wgdline (WGDInfo): WGDInfoLines that contains the styling information
    """
    TextCircle(
        ax,
        xx,
        yy,
        wgdline.name,
        fc=wgdline.color,
        radius=0.0225,
        color="k",
        fontweight="bold",
    )


def draw_wgd(ax, y, rescale, name, wgdcache):
    """Draw WGD given a name and the WGDInfo cache.

    Args:
        ax (matplotlib.axes): matplotlib axes
        y (float): y position
        rescale (function): Rescale function to generate x position
        name (str): Name of the line (usually the taxon/internal name)
        wgdcache (Dict): Dictionary containing WGDInfoLines
    """
    if not wgdcache or name not in wgdcache:
        return
    for line in wgdcache[name]:
        draw_wgd_xy(ax, rescale(line.divergence), y, line)


def draw_tree(
    ax,
    t,
    hpd=None,
    margin=0.1,
    rmargin=0.2,
    ymargin=0.1,
    tip=0.01,
    treecolor="k",
    supportcolor="k",
    internal=True,
    outgroup=None,
    dashedoutgroup=False,
    reroot=True,
    gffdir=None,
    sizes=None,
    trunc_name=None,
    SH=None,
    scutoff=0,
    leafcolor="k",
    leaffont=12,
    leafinfo=None,
    wgdinfo=None,
    geoscale=False,
    groups=[],
):
    """
    main function for drawing phylogenetic tree
    """

    if reroot:
        if outgroup:
            R = t.get_common_ancestor(*outgroup)
        else:
            # Calculate the midpoint node
            R = t.get_midpoint_outgroup()

        if R is not t:
            t.set_outgroup(R)

        # By default, the distance to outgroup and non-outgroup is the same
        # we re-adjust the distances so that the outgroups will appear
        # farthest from everything else
        if dashedoutgroup:
            a, b = t.children
            # Avoid even split
            total = a.dist + b.dist
            newR = t.get_common_ancestor(*outgroup)
            a.dist = 0.9 * total
            b.dist = total - a.dist

    farthest, max_dist = t.get_farthest_leaf()
    print("max_dist = {}".format(max_dist), file=sys.stderr)

    xstart = margin
    ystart = 2 * ymargin
    # scale the tree
    scale = (1 - margin - rmargin) / max_dist

    def rescale(dist):
        return xstart + scale * dist

    def rescale_divergence(divergence):
        return rescale(max_dist - divergence)

    num_leaves = len(t.get_leaf_names())
    yinterval = (1 - ystart) / num_leaves
    ytop = ystart + (num_leaves - 0.5) * yinterval

    # get exons structures, if any
    structures = {}
    if gffdir:
        gffiles = glob("{0}/*.gff*".format(gffdir))
        setups, ratio = get_setups(gffiles, canvas=rmargin / 2, noUTR=True)
        structures = dict((a, (b, c)) for a, b, c in setups)

    if sizes:
        sizes = Sizes(sizes).mapping

    coords = {}
    i = 0
    color_groups = []  # Used to plot groups to the right of the tree
    for n in t.traverse("postorder"):
        dist = n.get_distance(t)
        xx = rescale(dist)

        if n.is_leaf():
            yy = ystart + i * yinterval
            i += 1

            if trunc_name:
                name = truncate_name(n.name, rule=trunc_name)
            else:
                name = n.name

            if leafinfo and n.name in leafinfo:
                line = leafinfo[n.name]
                lc = line.color
                sname = line.new_name
            else:
                lc = leafcolor
                sname = None
            lc = lc or "k"
            sname = sname or name.replace("_", "-")
            # if color is given as "R,G,B"
            if "," in lc:
                lc = [float(x) for x in lc.split(",")]

            ax.text(
                xx + tip,
                yy,
                markup(sname),
                va="center",
                fontstyle="italic",
                size=leaffont,
                color=lc,
            )
            color_groups.append((lc, yy, xx))

            gname = n.name.split("_")[0]
            if gname in structures:
                mrnabed, cdsbeds = structures[gname]
                ExonGlyph(
                    ax,
                    1 - rmargin / 2,
                    yy,
                    mrnabed,
                    cdsbeds,
                    align="right",
                    ratio=ratio,
                )
            if sizes and gname in sizes:
                size = sizes[gname]
                size = size / 3 - 1  # base pair converted to amino acid
                size = "{0}aa".format(size)
                ax.text(1 - rmargin / 2 + tip, yy, size, size=leaffont)

        else:
            linestyle = "--" if (dashedoutgroup and n is t) else "-"
            children = [coords[x] for x in n.get_children()]
            children_x, children_y = zip(*children)
            min_y, max_y = min(children_y), max(children_y)
            # plot the vertical bar
            ax.plot((xx, xx), (min_y, max_y), linestyle, color=treecolor)
            # plot the horizontal bar
            for cx, cy in children:
                ax.plot((xx, cx), (cy, cy), linestyle, color=treecolor)
            yy = sum(children_y) * 1.0 / len(children_y)
            # plot HPD if exists
            if hpd and n.name in hpd:
                a, b = hpd[n.name]
                ax.plot(
                    (rescale_divergence(a), rescale_divergence(b)),
                    (yy, yy),
                    "-",
                    color="darkslategray",
                    alpha=0.4,
                    lw=2,
                )
            support = n.support
            if support > 1:
                support = support / 100.0
            if not n.is_root() and supportcolor:
                if support > scutoff / 100.0:
                    ax.text(
                        xx,
                        yy + 0.005,
                        "{0:d}".format(int(abs(support * 100))),
                        ha="right",
                        size=leaffont,
                        color=supportcolor,
                    )
            if internal and n.name:
                TextCircle(ax, xx, yy, n.name, size=9)
            else:  # Just a dot
                TextCircle(ax, xx, yy, None, radius=0.002)

        coords[n] = (xx, yy)
        # WGD info
        draw_wgd(ax, yy, rescale_divergence, n.name, wgdinfo)

    # scale bar
    if geoscale:
        draw_geoscale(
            ax, ytop, margin=margin, rmargin=rmargin, yy=ymargin, max_dist=max_dist
        )
    else:
        br = 0.1
        x1 = xstart + 0.1
        x2 = x1 + br * scale
        yy = ymargin
        ax.plot([x1, x1], [yy - tip, yy + tip], "-", color=treecolor)
        ax.plot([x2, x2], [yy - tip, yy + tip], "-", color=treecolor)
        ax.plot([x1, x2], [yy, yy], "-", color=treecolor)
        ax.text(
            (x1 + x2) / 2,
            yy - tip,
            "{0:g}".format(br),
            va="top",
            ha="center",
            size=leaffont,
            color=treecolor,
        )

    # Groupings on the right, often to used to show groups such as phylogenetic
    # clades
    if groups:
        color_groups.sort()
        group_extents = []
        for color, group in groupby(color_groups, key=lambda x: x[0]):
            group = list(group)
            _, min_yy, xx = min(group)
            _, max_yy, xx = max(group)
            group_extents.append((min_yy, max_yy, xx, color))
        group_extents.sort(reverse=True)

        for group_name, (min_yy, max_yy, xx, color) in zip(groups, group_extents):
            group_color = linear_shade(color, fraction=0.85)
            ax.add_patch(
                FancyBboxPatch(
                    (xx, min_yy - yinterval / 2),
                    rmargin - 0.01,
                    max_yy - min_yy + yinterval,
                    boxstyle="round,pad=-0.002,rounding_size=0.005",
                    fc=group_color,
                    ec=group_color,
                )
            )
            # Add the group label
            horizontal = (max_yy - min_yy) < 0.2
            mid_yy = (min_yy + max_yy) / 2
            label_rightend = 0.98
            if horizontal:
                ax.text(
                    label_rightend,
                    mid_yy,
                    markup(group_name),
                    color="darkslategray",
                    ha="right",
                    va="center",
                )
            else:
                ax.text(
                    label_rightend,
                    mid_yy,
                    markup(group_name),
                    color="darkslategray",
                    ha="right",
                    va="center",
                    rotation=-90,
                )

    if SH is not None:
        xs = x1
        ys = (ymargin + yy) / 2.0
        ax.text(
            xs,
            ys,
            "SH test against ref tree: {0}".format(SH),
            ha="left",
            size=leaffont,
            color="g",
        )


def read_trees(tree):
    from urllib.parse import parse_qs
    from jcvi.formats.base import read_block

    trees = []

    fp = open(tree)
    for header, tx in read_block(fp, "#"):
        header = parse_qs(header[1:])
        label = header["label"][0].strip('"')
        outgroup = header["outgroup"]
        (color,) = header.get("color", ["k"])
        trees.append((label, outgroup, color, "".join(tx)))

    return trees


def draw_geoscale(
    ax, ytop, margin=0.1, rmargin=0.2, yy=0.1, max_dist=3.0, contrast_epochs=True
):
    """
    Draw geological epoch on million year ago (mya) scale.
    max_dist = 3.0 => max is 300 mya
    """
    import math

    a, b = margin, 1 - rmargin  # Correspond to 300mya and 0mya
    minx, maxx = 0, int(max_dist * 100)

    def cv(x):
        return b - (x - b) / (maxx - minx) * (b - a)

    ax.plot((a, b), (yy, yy), "k-")
    tick = 0.0125
    scale_start = int(math.ceil(maxx / 25) * 25)
    for mya in range(scale_start - 25, 0, -25):
        p = cv(mya)
        ax.plot((p, p), (yy, yy - tick), "k-")
        ax.text(p, yy - 2.5 * tick, str(mya), ha="center", va="center")

    ax.text(
        (a + b) / 2,
        yy - 5 * tick,
        "Time before present (million years)",
        ha="center",
        va="center",
    )

    # Source:
    # https://en.wikipedia.org/wiki/Geological_period
    Geo = (
        ("Neogene", 2.588, 23.03),
        ("Paleogene", 23.03, 66.0),
        ("Cretaceous", 66.0, 145.5),
        ("Jurassic", 145.5, 201.3),
        ("Triassic", 201.3, 252.17),
        ("Permian", 252.17, 298.9),
        ("Carboniferous", 298.9, 358.9),
    )
    h = 0.05
    for (era, start, end), color in zip(Geo, set3_n(len(Geo))):
        if maxx - start < 10:  # not visible enough
            continue
        start, end = cv(start), cv(end)
        end = max(a, end)
        p = Rectangle((end, yy + tick / 2), abs(start - end), h, lw=1, ec="w", fc=color)
        ax.text(
            (start + end) / 2,
            yy + (tick + h) / 2,
            era,
            ha="center",
            va="center",
            size=8,
        )
        ax.add_patch(p)

    # We highlight recent epochs for better visualization, we just highlight
    # Neogene and Cretaceous as these are more relevant for most phylogeny
    if contrast_epochs:
        for era, start, end in Geo:
            if not era in ("Neogene", "Cretaceous"):
                continue

            # Make a beige patch
            start, end = cv(start), cv(end)
            ax.add_patch(
                Rectangle(
                    (end, yy + tick + h),
                    abs(start - end),
                    ytop - yy - tick - h,
                    fc="beige",
                    ec="beige",
                )
            )


def parse_tree(infile):
    """Parse newick formatted tree file and returns a tuple consisted of a
    Tree object, and a HPD dictionary if 95%HPD is found in the newick string,
    otherwise None

    Args:
        infile (str): Path to the tree file
    """
    import re

    with open(infile) as fp:
        treedata = fp.read()
    hpd_re = re.compile(r"( \[&95%HPD=[^[]*\])")

    def repl(match):
        repl.count += 1
        name = "N{}".format(repl.count)
        lb, ub = re.findall(r"HPD=\{(.*), (.*)\}", match.group(0))[0]
        repl.hpd[name] = (float(lb), float(ub))
        return name

    repl.count = 0
    repl.hpd = {}

    treedata, changed = re.subn(hpd_re, repl, treedata)
    if repl.hpd:
        print(repl.hpd, file=sys.stderr)

    return (Tree(treedata, format=1), repl.hpd) if changed else (Tree(treedata), None)


def main(args):
    """
    %prog newicktree

    Plot Newick formatted tree. The gene structure can be plotted along if
    --gffdir is given. The gff file needs to be `genename.gff`. If --sizes is
    on, also show the number of amino acids.
    """
    p = OptionParser(main.__doc__)
    p.add_option(
        "--outgroup",
        help="Outgroup for rerooting the tree. "
        + "Use comma to separate multiple taxa.",
    )
    p.add_option(
        "--noreroot",
        default=False,
        action="store_true",
        help="Don't reroot the input tree",
    )
    p.add_option(
        "--rmargin", default=0.2, type="float", help="Set blank rmargin to the right"
    )
    p.add_option("--gffdir", default=None, help="The directory that contain GFF files")
    p.add_option("--sizes", default=None, help="The FASTA file or the sizes file")
    p.add_option("--SH", default=None, type="string", help="SH test p-value")

    group = p.add_option_group("Node style")
    group.add_option("--leafcolor", default="k", help="Font color for the OTUs")
    group.add_option("--leaffont", default=12, help="Font size for the OTUs")
    group.add_option(
        "--leafinfo", help="CSV file specifying the leaves: name,color,new_name"
    )
    group.add_option(
        "--scutoff",
        default=0,
        type="int",
        help="cutoff for displaying node support, 0-100",
    )
    group.add_option(
        "--no_support",
        dest="support",
        default=True,
        action="store_false",
        help="Do not print node support values",
    )
    group.add_option(
        "--no_internal",
        dest="internal",
        default=True,
        action="store_false",
        help="Do not show internal nodes",
    )

    group = p.add_option_group("Edge style")
    group.add_option(
        "--dashedoutgroup",
        default=False,
        action="store_true",
        help="Gray out the edges connecting outgroup and non-outgroup",
    )

    group = p.add_option_group("Additional annotations")
    group.add_option(
        "--geoscale",
        default=False,
        action="store_true",
        help="Plot geological scale",
    )
    group.add_option(
        "--wgdinfo", help="CSV specifying the position and style of WGD events"
    )
    group.add_option(
        "--groups",
        help="Group names from top to bottom, to the right of the tree. "
        "Each distinct color in --leafinfo is considered part of the same group. "
        "Separate the names with comma, such as 'eudicots,,monocots,'. "
        "Empty names will be ignored for that specific group. ",
    )

    opts, args, iopts = p.set_image_options(args, figsize="10x7")

    if len(args) != 1:
        sys.exit(not p.print_help())

    (datafile,) = args
    outgroup = None
    reroot = not opts.noreroot
    if opts.outgroup:
        outgroup = opts.outgroup.split(",")

    hpd = None
    if datafile == "demo":
        t = Tree(
            """(((Os02g0681100:0.1151,Sb04g031800:0.11220)1.0:0.0537,
        (Os04g0578800:0.04318,Sb06g026210:0.04798)-1.0:0.08870)1.0:0.06985,
        ((Os03g0124100:0.08845,Sb01g048930:0.09055)1.0:0.05332,
        (Os10g0534700:0.06592,Sb01g030630:0.04824)-1.0:0.07886):0.09389);"""
        )
    else:
        logging.debug("Load tree file `{0}`".format(datafile))
        t, hpd = parse_tree(datafile)

    pf = datafile.rsplit(".", 1)[0]

    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])

    supportcolor = "k" if opts.support else None
    margin, rmargin = 0.1, opts.rmargin  # Left and right margin
    leafinfo = LeafInfoFile(opts.leafinfo).cache if opts.leafinfo else None
    wgdinfo = WGDInfoFile(opts.wgdinfo).cache if opts.wgdinfo else None

    draw_tree(
        root,
        t,
        hpd=hpd,
        margin=margin,
        rmargin=rmargin,
        ymargin=margin,
        supportcolor=supportcolor,
        internal=opts.internal,
        outgroup=outgroup,
        dashedoutgroup=opts.dashedoutgroup,
        reroot=reroot,
        gffdir=opts.gffdir,
        sizes=opts.sizes,
        SH=opts.SH,
        scutoff=opts.scutoff,
        leafcolor=opts.leafcolor,
        leaffont=opts.leaffont,
        leafinfo=leafinfo,
        wgdinfo=wgdinfo,
        geoscale=opts.geoscale,
        groups=opts.groups.split(",") if opts.groups else [],
    )

    normalize_axes(root)

    image_name = pf + "." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


if __name__ == "__main__":
    main(sys.argv[1:])
